# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from typing import List
from typing import Any
from typing import Dict

import os
import abc
import six
import json
import logging

from functools import wraps

import yaml
from six.moves import http_client

from st2client import commands
from st2client.exceptions.operations import OperationFailureException
from st2client.formatters import table
from st2client.utils.types import OrderedSet

ALLOWED_EXTS = [".json", ".yaml", ".yml"]
PARSER_FUNCS = {".json": json.load, ".yml": yaml.safe_load, ".yaml": yaml.safe_load}
LOG = logging.getLogger(__name__)


def add_auth_token_to_kwargs_from_cli(func):
    @wraps(func)
    def decorate(*args, **kwargs):
        ns = args[1]
        if getattr(ns, "token", None):
            kwargs["token"] = ns.token
        if getattr(ns, "api_key", None):
            kwargs["api_key"] = ns.api_key
        return func(*args, **kwargs)

    return decorate


class ResourceCommandError(Exception):
    pass


class ResourceNotFoundError(Exception):
    pass


class ResourceBranch(commands.Branch):
    def __init__(
        self,
        resource,
        description,
        app,
        subparsers,
        parent_parser=None,
        read_only=False,
        commands=None,
        has_disable=False,
    ):

        self.resource = resource
        super(ResourceBranch, self).__init__(
            self.resource.get_alias().lower(),
            description,
            app,
            subparsers,
            parent_parser=parent_parser,
        )

        # Registers subcommands for managing the resource type.
        self.subparsers = self.parser.add_subparsers(
            help=(
                "List of commands for managing %s."
                % self.resource.get_plural_display_name().lower()
            )
        )

        # Resolves if commands need to be overridden.
        commands = commands or {}
        cmd_map = {
            "list": ResourceListCommand,
            "get": ResourceGetCommand,
            "create": ResourceCreateCommand,
            "update": ResourceUpdateCommand,
            "delete": ResourceDeleteCommand,
            "enable": ResourceEnableCommand,
            "disable": ResourceDisableCommand,
            "clone": ResourceCloneCommand,
        }
        for cmd, cmd_class in cmd_map.items():
            if cmd not in commands:
                commands[cmd] = cmd_class

        # Instantiate commands.
        args = [self.resource, self.app, self.subparsers]
        self.commands["list"] = commands["list"](*args)
        self.commands["get"] = commands["get"](*args)

        if not read_only:
            self.commands["create"] = commands["create"](*args)
            self.commands["update"] = commands["update"](*args)
            self.commands["delete"] = commands["delete"](*args)
            self.commands["clone"] = commands["clone"](*args)

        if has_disable:
            self.commands["enable"] = commands["enable"](*args)
            self.commands["disable"] = commands["disable"](*args)


@six.add_metaclass(abc.ABCMeta)
class ResourceCommand(commands.Command):
    pk_argument_name = None

    def __init__(self, resource, *args, **kwargs):

        has_token_opt = kwargs.pop("has_token_opt", True)

        super(ResourceCommand, self).__init__(*args, **kwargs)

        self.resource = resource

        if has_token_opt:
            self.parser.add_argument(
                "-t",
                "--token",
                dest="token",
                help="Access token for user authentication. "
                "Get ST2_AUTH_TOKEN from the environment "
                "variables by default.",
            )
            self.parser.add_argument(
                "--api-key",
                dest="api_key",
                help="Api Key for user authentication. "
                "Get ST2_API_KEY from the environment "
                "variables by default.",
            )

        # Formatter flags
        self.parser.add_argument(
            "-j",
            "--json",
            action="store_true",
            dest="json",
            help="Print output in JSON format.",
        )
        self.parser.add_argument(
            "-y",
            "--yaml",
            action="store_true",
            dest="yaml",
            help="Print output in YAML format.",
        )

    @property
    def manager(self):
        return self.app.client.managers[self.resource.__name__]

    @property
    def arg_name_for_resource_id(self):
        resource_name = self.resource.get_display_name().lower()
        return "%s-id" % resource_name.replace(" ", "-")

    def print_not_found(self, name):
        print('%s "%s" is not found.\n' % (self.resource.get_display_name(), name))

    def get_resource(self, name_or_id, **kwargs):
        pk_argument_name = self.pk_argument_name

        if pk_argument_name == "name_or_id":
            instance = self.get_resource_by_name_or_id(name_or_id=name_or_id, **kwargs)
        elif pk_argument_name == "ref_or_id":
            instance = self.get_resource_by_ref_or_id(ref_or_id=name_or_id, **kwargs)
        else:
            instance = self.get_resource_by_pk(pk=name_or_id, **kwargs)

        return instance

    def get_resource_by_pk(self, pk, **kwargs):
        """
        Retrieve resource by a primary key.
        """
        try:
            instance = self.manager.get_by_id(pk, **kwargs)
        except Exception as e:
            # Hack for "Unauthorized" exceptions, we do want to propagate those
            response = getattr(e, "response", None)
            status_code = getattr(response, "status_code", None)
            if status_code and status_code == http_client.FORBIDDEN:
                raise e
            if status_code and status_code == http_client.UNAUTHORIZED:
                raise e

            instance = None

        return instance

    def get_resource_by_id(self, id, **kwargs):
        instance = self.get_resource_by_pk(pk=id, **kwargs)

        if not instance:
            message = 'Resource with id "%s" doesn\'t exist.' % (id)
            raise ResourceNotFoundError(message)
        return instance

    def get_resource_by_name(self, name, **kwargs):
        """
        Retrieve resource by name.
        """
        instance = self.manager.get_by_name(name, **kwargs)
        return instance

    def get_resource_by_name_or_id(self, name_or_id, **kwargs):
        instance = self.get_resource_by_name(name=name_or_id, **kwargs)
        if not instance:
            instance = self.get_resource_by_pk(pk=name_or_id, **kwargs)

        if not instance:
            message = 'Resource with id or name "%s" doesn\'t exist.' % (name_or_id)
            raise ResourceNotFoundError(message)
        return instance

    def get_resource_by_ref_or_id(self, ref_or_id, **kwargs):
        instance = self.manager.get_by_ref_or_id(ref_or_id=ref_or_id, **kwargs)

        if not instance:
            message = 'Resource with id or reference "%s" doesn\'t exist.' % (ref_or_id)
            raise ResourceNotFoundError(message)
        return instance

    def _get_multiple_resources(
        self, resource_ids: List[str], kwargs: Dict[str, Any]
    ) -> List[Any]:
        """
        Return multiple resource instances for the provided resource ids.
        If a resource is not found, an error is printed. This method only throws when operating on
        a single resource.
        :param resource_ids: A list of resources to retrieve instances for.
        :param kwargs: Dictionary with keyword arguments which are passed to get_resource_by_id.
        """
        more_than_one_resource = len(resource_ids) > 1

        resources = []
        for resource_id in resource_ids:
            try:
                resource = self.get_resource_by_id(resource_id, **kwargs)
            except ResourceNotFoundError:
                self.print_not_found(resource_id)

                if not more_than_one_resource:
                    # For backward compatibility reasons and to comply with common "get one"
                    # behavior, we only fail if a single source is requested
                    raise ResourceNotFoundError("Resource %s not found." % resource_id)

                continue

            resources.append(resource)
        return resources

    @abc.abstractmethod
    def run(self, args, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def run_and_print(self, args, **kwargs):
        raise NotImplementedError

    def _get_metavar_for_argument(self, argument):
        return argument.replace("_", "-")

    def _get_help_for_argument(self, resource, argument):
        argument_display_name = argument.title()
        resource_display_name = resource.get_display_name().lower()

        if "ref" in argument:
            result = "Reference or ID of the %s." % (resource_display_name)
        elif "name_or_id" in argument:
            result = "Name or ID of the %s." % (resource_display_name)
        else:
            result = "%s of the %s." % (argument_display_name, resource_display_name)

        return result


class ResourceViewCommand(ResourceCommand):
    """
    Base class for read / view commands (list and get).
    """

    @classmethod
    def _get_include_attributes(cls, args, extra_attributes=None):
        """
        Return a list of attributes to send to the API using ?include_attributes filter.

        If None / empty list is returned it's assumed no filtering is to be performed and all
        attributes are to be retrieved.

        :param extra_attributes: Additional include attributes which should always be included.
        :type extra_attributes: ``list`` of ``str``
        """
        extra_attributes = extra_attributes or []

        include_attributes = []

        if extra_attributes:
            include_attributes.extend(extra_attributes)

        # If user specifies which attributes to retrieve via CLI --attr / -a argument, take that
        # into account

        # Special case for "all"
        if "all" in args.attr:
            return None

        for attr in args.attr:
            include_attributes.append(attr)

        if include_attributes:
            return include_attributes

        display_attributes = getattr(cls, "display_attributes", [])

        if display_attributes:
            include_attributes += display_attributes

        include_attributes = list(OrderedSet(include_attributes))

        return include_attributes


class ResourceTableCommand(ResourceViewCommand):
    display_attributes = ["id", "name", "description"]

    def __init__(self, resource, name, description, *args, **kwargs):
        super(ResourceTableCommand, self).__init__(
            resource, name, description, *args, **kwargs
        )

        self.parser.add_argument(
            "-a",
            "--attr",
            nargs="+",
            default=self.display_attributes,
            help=(
                "List of attributes to include in the "
                'output. "all" will return all '
                "attributes."
            ),
        )
        self.parser.add_argument(
            "-w",
            "--width",
            nargs="+",
            type=int,
            default=None,
            help=("Set the width of columns in output."),
        )

    @add_auth_token_to_kwargs_from_cli
    def run(self, args, **kwargs):
        include_attributes = self._get_include_attributes(args=args)
        if include_attributes:
            include_attributes = ",".join(include_attributes)
            kwargs["params"] = {"include_attributes": include_attributes}

        return self.manager.get_all(**kwargs)

    def run_and_print(self, args, **kwargs):
        instances = self.run(args, **kwargs)
        self.print_output(
            instances,
            table.MultiColumnTable,
            attributes=args.attr,
            widths=args.width,
            json=args.json,
            yaml=args.yaml,
        )


class ResourceListCommand(ResourceTableCommand):
    def __init__(self, resource, *args, **kwargs):
        super(ResourceListCommand, self).__init__(
            resource,
            "list",
            "Get the list of %s." % resource.get_plural_display_name().lower(),
            *args,
            **kwargs,
        )


class ContentPackResourceListCommand(ResourceListCommand):
    """
    Base command class for use with resources which belong to a content pack.
    """

    def __init__(self, resource, *args, **kwargs):
        super(ContentPackResourceListCommand, self).__init__(resource, *args, **kwargs)

        self.parser.add_argument(
            "-p",
            "--pack",
            type=str,
            help=("Only return resources belonging to the" " provided pack"),
        )

    @add_auth_token_to_kwargs_from_cli
    def run(self, args, **kwargs):
        filters = {"pack": args.pack}
        filters.update(**kwargs)

        include_attributes = self._get_include_attributes(args=args)
        if include_attributes:
            include_attributes = ",".join(include_attributes)
            filters["params"] = {"include_attributes": include_attributes}

        return self.manager.get_all(**filters)


class ResourceGetCommand(ResourceViewCommand):
    display_attributes = ["all"]
    attribute_display_order = ["id", "name", "description"]

    pk_argument_name = "name_or_id"  # name of the attribute which stores resource PK

    help_string = None

    def __init__(self, resource, *args, **kwargs):
        super(ResourceGetCommand, self).__init__(
            resource,
            "get",
            self.help_string
            or "Get individual %s." % resource.get_display_name().lower(),
            *args,
            **kwargs,
        )

        argument = self.pk_argument_name
        metavar = self._get_metavar_for_argument(argument=self.pk_argument_name)
        help = self._get_help_for_argument(
            resource=resource, argument=self.pk_argument_name
        )

        self.parser.add_argument(argument, metavar=metavar, nargs="+", help=help)
        self.parser.add_argument(
            "-a",
            "--attr",
            nargs="+",
            default=self.display_attributes,
            help=(
                "List of attributes to include in the "
                'output. "all" or unspecified will '
                "return all attributes."
            ),
        )

    @add_auth_token_to_kwargs_from_cli
    def run(self, args, **kwargs):
        resource_ids = getattr(args, self.pk_argument_name, None)
        resources = self._get_multiple_resources(
            resource_ids=resource_ids, kwargs=kwargs
        )
        return resources

    def run_and_print(self, args, **kwargs):
        instances = self.run(args, **kwargs)

        for instance in instances:
            self.print_output(
                instance,
                table.PropertyValueTable,
                attributes=args.attr,
                json=args.json,
                yaml=args.yaml,
                attribute_display_order=self.attribute_display_order,
            )


class ContentPackResourceGetCommand(ResourceGetCommand):
    """
    Command for retrieving a single resource which belongs to a content pack.

    Note: All the resources which belong to the content pack can either be
    retrieved by a reference or by an id.
    """

    attribute_display_order = ["id", "pack", "name", "description"]

    pk_argument_name = "ref_or_id"

    def get_resource(self, ref_or_id, **kwargs):
        return self.get_resource_by_ref_or_id(ref_or_id=ref_or_id, **kwargs)


class ResourceCreateCommand(ResourceCommand):
    def __init__(self, resource, *args, **kwargs):
        super(ResourceCreateCommand, self).__init__(
            resource,
            "create",
            "Create a new %s." % resource.get_display_name().lower(),
            *args,
            **kwargs,
        )

        self.parser.add_argument(
            "file",
            help=(
                "JSON/YAML file containing the %s to create."
                % resource.get_display_name().lower()
            ),
        )

    @add_auth_token_to_kwargs_from_cli
    def run(self, args, **kwargs):
        data = load_meta_file(args.file)
        instance = self.resource.deserialize(data)
        return self.manager.create(instance, **kwargs)

    def run_and_print(self, args, **kwargs):
        try:
            instance = self.run(args, **kwargs)
            if not instance:
                raise Exception("Server did not create instance.")
            self.print_output(
                instance,
                table.PropertyValueTable,
                attributes=["all"],
                json=args.json,
                yaml=args.yaml,
            )
        except Exception as e:
            message = six.text_type(e)
            print("ERROR: %s" % (message))
            raise OperationFailureException(message)


class ResourceUpdateCommand(ResourceCommand):
    pk_argument_name = "name_or_id"

    def __init__(self, resource, *args, **kwargs):
        super(ResourceUpdateCommand, self).__init__(
            resource,
            "update",
            "Updating an existing %s." % resource.get_display_name().lower(),
            *args,
            **kwargs,
        )

        argument = self.pk_argument_name
        metavar = self._get_metavar_for_argument(argument=self.pk_argument_name)
        help = self._get_help_for_argument(
            resource=resource, argument=self.pk_argument_name
        )

        self.parser.add_argument(argument, metavar=metavar, help=help)
        self.parser.add_argument(
            "file",
            help=(
                "JSON/YAML file containing the %s to update."
                % resource.get_display_name().lower()
            ),
        )

    @add_auth_token_to_kwargs_from_cli
    def run(self, args, **kwargs):
        resource_id = getattr(args, self.pk_argument_name, None)
        instance = self.get_resource(resource_id, **kwargs)
        data = load_meta_file(args.file)
        modified_instance = self.resource.deserialize(data)

        if not getattr(modified_instance, "id", None):
            modified_instance.id = instance.id
        else:
            if modified_instance.id != instance.id:
                raise Exception(
                    "The value for the %s id in the JSON/YAML file "
                    "does not match the ID provided in the "
                    "command line arguments." % self.resource.get_display_name().lower()
                )
        return self.manager.update(modified_instance, **kwargs)

    def run_and_print(self, args, **kwargs):
        instance = self.run(args, **kwargs)
        try:
            self.print_output(
                instance,
                table.PropertyValueTable,
                attributes=["all"],
                json=args.json,
                yaml=args.yaml,
            )
        except Exception as e:
            print("ERROR: %s" % (six.text_type(e)))
            raise OperationFailureException(six.text_type(e))


class ContentPackResourceUpdateCommand(ResourceUpdateCommand):
    pk_argument_name = "ref_or_id"


class ResourceEnableCommand(ResourceCommand):
    pk_argument_name = "name_or_id"

    def __init__(self, resource, *args, **kwargs):
        super(ResourceEnableCommand, self).__init__(
            resource,
            "enable",
            "Enable an existing %s." % resource.get_display_name().lower(),
            *args,
            **kwargs,
        )

        argument = self.pk_argument_name
        metavar = self._get_metavar_for_argument(argument=self.pk_argument_name)
        help = self._get_help_for_argument(
            resource=resource, argument=self.pk_argument_name
        )

        self.parser.add_argument(argument, metavar=metavar, help=help)

    @add_auth_token_to_kwargs_from_cli
    def run(self, args, **kwargs):
        resource_id = getattr(args, self.pk_argument_name, None)
        instance = self.get_resource(resource_id, **kwargs)

        data = instance.serialize()

        if "ref" in data:
            del data["ref"]

        data["enabled"] = True
        modified_instance = self.resource.deserialize(data)

        return self.manager.update(modified_instance, **kwargs)

    def run_and_print(self, args, **kwargs):
        instance = self.run(args, **kwargs)
        self.print_output(
            instance,
            table.PropertyValueTable,
            attributes=["all"],
            json=args.json,
            yaml=args.yaml,
        )


class ContentPackResourceEnableCommand(ResourceEnableCommand):
    pk_argument_name = "ref_or_id"


class ResourceDisableCommand(ResourceCommand):
    pk_argument_name = "name_or_id"

    def __init__(self, resource, *args, **kwargs):
        super(ResourceDisableCommand, self).__init__(
            resource,
            "disable",
            "Disable an existing %s." % resource.get_display_name().lower(),
            *args,
            **kwargs,
        )

        argument = self.pk_argument_name
        metavar = self._get_metavar_for_argument(argument=self.pk_argument_name)
        help = self._get_help_for_argument(
            resource=resource, argument=self.pk_argument_name
        )

        self.parser.add_argument(argument, metavar=metavar, help=help)

    @add_auth_token_to_kwargs_from_cli
    def run(self, args, **kwargs):
        resource_id = getattr(args, self.pk_argument_name, None)
        instance = self.get_resource(resource_id, **kwargs)

        data = instance.serialize()

        if "ref" in data:
            del data["ref"]

        data["enabled"] = False
        modified_instance = self.resource.deserialize(data)

        return self.manager.update(modified_instance, **kwargs)

    def run_and_print(self, args, **kwargs):
        instance = self.run(args, **kwargs)
        self.print_output(
            instance,
            table.PropertyValueTable,
            attributes=["all"],
            json=args.json,
            yaml=args.yaml,
        )


class ContentPackResourceDisableCommand(ResourceDisableCommand):
    pk_argument_name = "ref_or_id"


class ResourceDeleteCommand(ResourceCommand):
    pk_argument_name = "name_or_id"

    def __init__(self, resource, *args, **kwargs):
        super(ResourceDeleteCommand, self).__init__(
            resource,
            "delete",
            "Delete an existing %s." % resource.get_display_name().lower(),
            *args,
            **kwargs,
        )

        argument = self.pk_argument_name
        metavar = self._get_metavar_for_argument(argument=self.pk_argument_name)
        help = self._get_help_for_argument(
            resource=resource, argument=self.pk_argument_name
        )

        self.parser.add_argument(argument, metavar=metavar, help=help)

    @add_auth_token_to_kwargs_from_cli
    def run(self, args, **kwargs):
        resource_id = getattr(args, self.pk_argument_name, None)
        instance = self.get_resource(resource_id, **kwargs)
        self.manager.delete(instance, **kwargs)

    def run_and_print(self, args, **kwargs):
        resource_id = getattr(args, self.pk_argument_name, None)

        try:
            self.run(args, **kwargs)
            print(
                'Resource with id "%s" has been successfully deleted.' % (resource_id)
            )
        except ResourceNotFoundError:
            self.print_not_found(resource_id)
            raise OperationFailureException("Resource %s not found." % resource_id)


class ContentPackResourceDeleteCommand(ResourceDeleteCommand):
    """
    Base command class for deleting a resource which belongs to a content pack.
    """

    pk_argument_name = "ref_or_id"


class ResourceCloneCommand(ResourceCommand):
    def __init__(self, resource, *args, **kwargs):
        super(ResourceCloneCommand, self).__init__(
            resource,
            "clone",
            "Clone a new %s." % resource.get_display_name().lower(),
            *args,
            **kwargs,
        )

    @add_auth_token_to_kwargs_from_cli
    def run(self, args, **kwargs):
        raise NotImplementedError("clone '%s' is not implemented." % args.parser)

    def run_and_print(self, args, **kwargs):
        self.run(args, **kwargs)


class ContentPackResourceCloneCommand(ResourceCloneCommand):
    pass


def load_meta_file(file_path):
    if not os.path.isfile(file_path):
        raise Exception('File "%s" does not exist.' % file_path)

    file_name, file_ext = os.path.splitext(file_path)
    if file_ext not in ALLOWED_EXTS:
        raise Exception(
            "Unsupported meta type %s, file %s. Allowed: %s"
            % (file_ext, file_path, ALLOWED_EXTS)
        )

    with open(file_path, "r") as f:
        return PARSER_FUNCS[file_ext](f)



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import argparse
import json
import logging
import mock
import os
import requests
import tempfile
import yaml

from tests import base
from st2client import shell

LOG = logging.getLogger(__name__)

KEYVALUE = {
    "id": "kv_name",
    "name": "kv_name.",
    "value": "super cool value",
    "scope": "system",
}

KEYVALUE_USER = {
    "id": "kv_name",
    "name": "kv_name.",
    "value": "super cool value",
    "scope": "system",
    "user": "stanley",
}

KEYVALUE_SECRET = {
    "id": "kv_name",
    "name": "kv_name.",
    "value": "super cool value",
    "scope": "system",
    "secret": True,
}

KEYVALUE_PRE_ENCRYPTED = {
    "id": "kv_name",
    "name": "kv_name.",
    "value": "AAABBBCCC1234",
    "scope": "system",
    "encrypted": True,
    "secret": True,
}

KEYVALUE_TTL = {
    "id": "kv_name",
    "name": "kv_name.",
    "value": "super cool value",
    "scope": "system",
    "ttl": 100,
}

KEYVALUE_OBJECT = {
    "id": "kv_name",
    "name": "kv_name.",
    "value": {"obj": [1, True, 23.4, "abc"]},
    "scope": "system",
}

KEYVALUE_ALL = {
    "id": "kv_name",
    "name": "kv_name.",
    "value": "AAAAABBBBBCCCCCCDDDDD11122345",
    "scope": "system",
    "user": "stanley",
    "secret": True,
    "encrypted": True,
    "ttl": 100,
}

KEYVALUE_MISSING_NAME = {"id": "kv_name", "value": "super cool value"}

KEYVALUE_MISSING_VALUE = {"id": "kv_name", "name": "kv_name."}


class TestKeyValueBase(base.BaseCLITestCase):
    """Base class for "key" CLI tests"""

    capture_output = True

    def __init__(self, *args, **kwargs):
        super(TestKeyValueBase, self).__init__(*args, **kwargs)

        self.parser = argparse.ArgumentParser()
        self.parser.add_argument("-t", "--token", dest="token")
        self.parser.add_argument("--api-key", dest="api_key")
        self.shell = shell.Shell()

    def setUp(self):
        super(TestKeyValueBase, self).setUp()

    def tearDown(self):
        super(TestKeyValueBase, self).tearDown()


class TestKeyValueSet(TestKeyValueBase):
    @mock.patch.object(
        requests,
        "put",
        mock.MagicMock(
            return_value=base.FakeResponse(
                json.dumps(KEYVALUE_PRE_ENCRYPTED), 200, "OK"
            )
        ),
    )
    def test_set_keyvalue(self):
        """Test setting key/value pair with optional pre_encrypted field"""
        args = ["key", "set", "--encrypted", "kv_name", "AAABBBCCC1234"]
        retcode = self.shell.run(args)
        self.assertEqual(retcode, 0)

    def test_encrypt_and_encrypted_flags_are_mutually_exclusive(self):
        args = ["key", "set", "--encrypt", "--encrypted", "kv_name", "AAABBBCCC1234"]

        self.assertRaisesRegexp(SystemExit, "2", self.shell.run, args)

        self.stderr.seek(0)
        stderr = self.stderr.read()

        expected_msg = (
            "error: argument --encrypted: not allowed with argument -e/--encrypt"
        )
        self.assertIn(expected_msg, stderr)


class TestKeyValueLoad(TestKeyValueBase):
    @mock.patch.object(
        requests,
        "put",
        mock.MagicMock(return_value=base.FakeResponse(json.dumps(KEYVALUE), 200, "OK")),
    )
    def test_load_keyvalue_json(self):
        """Test loading of key/value pair in JSON format"""
        fd, path = tempfile.mkstemp(suffix=".json")
        try:
            with open(path, "a") as f:
                f.write(json.dumps(KEYVALUE, indent=4))

            args = ["key", "load", path]
            retcode = self.shell.run(args)
            self.assertEqual(retcode, 0)
        finally:
            os.close(fd)
            os.unlink(path)

    @mock.patch.object(
        requests,
        "put",
        mock.MagicMock(return_value=base.FakeResponse(json.dumps(KEYVALUE), 200, "OK")),
    )
    def test_load_keyvalue_yaml(self):
        """Test loading of key/value pair in YAML format"""
        fd, path = tempfile.mkstemp(suffix=".yaml")
        try:
            with open(path, "a") as f:
                f.write(yaml.safe_dump(KEYVALUE))

            args = ["key", "load", path]
            retcode = self.shell.run(args)
            self.assertEqual(retcode, 0)
        finally:
            os.close(fd)
            os.unlink(path)

    @mock.patch.object(
        requests,
        "put",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(KEYVALUE_USER), 200, "OK")
        ),
    )
    def test_load_keyvalue_user(self):
        """Test loading of key/value pair with the optional user field"""
        fd, path = tempfile.mkstemp(suffix=".json")
        try:
            with open(path, "a") as f:
                f.write(json.dumps(KEYVALUE_USER, indent=4))

            args = ["key", "load", path]
            retcode = self.shell.run(args)
            self.assertEqual(retcode, 0)
        finally:
            os.close(fd)
            os.unlink(path)

    @mock.patch.object(
        requests,
        "put",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(KEYVALUE_SECRET), 200, "OK")
        ),
    )
    def test_load_keyvalue_secret(self):
        """Test loading of key/value pair with the optional secret field"""
        fd, path = tempfile.mkstemp(suffix=".json")
        try:
            with open(path, "a") as f:
                f.write(json.dumps(KEYVALUE_SECRET, indent=4))

            args = ["key", "load", path]
            retcode = self.shell.run(args)
            self.assertEqual(retcode, 0)
        finally:
            os.close(fd)
            os.unlink(path)

    @mock.patch.object(
        requests,
        "put",
        mock.MagicMock(
            return_value=base.FakeResponse(
                json.dumps(KEYVALUE_PRE_ENCRYPTED), 200, "OK"
            )
        ),
    )
    def test_load_keyvalue_already_encrypted(self):
        """Test loading of key/value pair with the pre-encrypted value"""
        fd, path = tempfile.mkstemp(suffix=".json")
        try:
            with open(path, "a") as f:
                f.write(json.dumps(KEYVALUE_PRE_ENCRYPTED, indent=4))

            args = ["key", "load", path]
            retcode = self.shell.run(args)
            self.assertEqual(retcode, 0)
        finally:
            os.close(fd)
            os.unlink(path)

    @mock.patch.object(
        requests,
        "put",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(KEYVALUE_TTL), 200, "OK")
        ),
    )
    def test_load_keyvalue_ttl(self):
        """Test loading of key/value pair with the optional ttl field"""
        fd, path = tempfile.mkstemp(suffix=".json")
        try:
            with open(path, "a") as f:
                f.write(json.dumps(KEYVALUE_TTL, indent=4))

            args = ["key", "load", path]
            retcode = self.shell.run(args)
            self.assertEqual(retcode, 0)
        finally:
            os.close(fd)
            os.unlink(path)

    @mock.patch.object(
        requests,
        "put",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(KEYVALUE_OBJECT), 200, "OK")
        ),
    )
    def test_load_keyvalue_object(self):
        """Test loading of key/value pair where the value is an object"""
        fd, path = tempfile.mkstemp(suffix=".json")
        try:
            with open(path, "a") as f:
                f.write(json.dumps(KEYVALUE_OBJECT, indent=4))

            # test converting with short option
            args = ["key", "load", "-c", path]
            retcode = self.shell.run(args)
            self.assertEqual(retcode, 0)

            # test converting with long option
            args = ["key", "load", "--convert", path]
            retcode = self.shell.run(args)
            self.assertEqual(retcode, 0)
        finally:
            os.close(fd)
            os.unlink(path)

    @mock.patch.object(
        requests,
        "put",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(KEYVALUE_OBJECT), 200, "OK")
        ),
    )
    def test_load_keyvalue_object_fail(self):
        """Test failure to load key/value pair where the value is an object
        and the -c/--convert option is not passed
        """
        fd, path = tempfile.mkstemp(suffix=".json")
        try:
            with open(path, "a") as f:
                f.write(json.dumps(KEYVALUE_OBJECT, indent=4))

            # test converting with short option
            args = ["key", "load", path]
            retcode = self.shell.run(args)
            self.assertNotEqual(retcode, 0)
        finally:
            os.close(fd)
            os.unlink(path)

    @mock.patch.object(
        requests,
        "put",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(KEYVALUE_ALL), 200, "OK")
        ),
    )
    def test_load_keyvalue_all(self):
        """Test loading of key/value pair with all optional fields"""
        fd, path = tempfile.mkstemp(suffix=".json")
        try:
            with open(path, "a") as f:
                f.write(json.dumps(KEYVALUE_ALL, indent=4))

            args = ["key", "load", path]
            retcode = self.shell.run(args)
            self.assertEqual(retcode, 0)
        finally:
            os.close(fd)
            os.unlink(path)

    @mock.patch.object(
        requests,
        "put",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(KEYVALUE_ALL), 200, "OK")
        ),
    )
    def test_load_keyvalue_array(self):
        """Test loading an array of key/value pairs"""
        fd, path = tempfile.mkstemp(suffix=".json")
        try:
            array = [KEYVALUE, KEYVALUE_ALL]
            json_str = json.dumps(array, indent=4)
            LOG.info(json_str)
            with open(path, "a") as f:
                f.write(json_str)

            args = ["key", "load", path]
            retcode = self.shell.run(args)
            self.assertEqual(retcode, 0)
        finally:
            os.close(fd)
            os.unlink(path)

    def test_load_keyvalue_missing_name(self):
        """Test loading of a key/value pair with the required field 'name' missing"""
        fd, path = tempfile.mkstemp(suffix=".json")
        try:
            with open(path, "a") as f:
                f.write(json.dumps(KEYVALUE_MISSING_NAME, indent=4))

            args = ["key", "load", path]
            retcode = self.shell.run(args)
            self.assertEqual(retcode, 1)
        finally:
            os.close(fd)
            os.unlink(path)

    def test_load_keyvalue_missing_value(self):
        """Test loading of a key/value pair with the required field 'value' missing"""
        fd, path = tempfile.mkstemp(suffix=".json")
        try:
            with open(path, "a") as f:
                f.write(json.dumps(KEYVALUE_MISSING_VALUE, indent=4))

            args = ["key", "load", path]
            retcode = self.shell.run(args)
            self.assertEqual(retcode, 1)
        finally:
            os.close(fd)
            os.unlink(path)

    def test_load_keyvalue_missing_file(self):
        """Test loading of a key/value pair with a missing file"""
        path = "/some/file/that/doesnt/exist.json"
        args = ["key", "load", path]
        retcode = self.shell.run(args)
        self.assertEqual(retcode, 1)

    def test_load_keyvalue_bad_file_extension(self):
        """Test loading of a key/value pair with a bad file extension"""
        fd, path = tempfile.mkstemp(suffix=".badext")
        try:
            args = ["key", "load", path]
            retcode = self.shell.run(args)
            self.assertEqual(retcode, 1)
        finally:
            os.close(fd)
            os.unlink(path)

    def test_load_keyvalue_empty_file(self):
        """
        Loading K/V from an empty file shouldn't throw an error
        """
        fd, path = tempfile.mkstemp(suffix=".yaml")
        try:
            args = ["key", "load", path]
            retcode = self.shell.run(args)
            self.assertIn("No matching items found", self.stdout.getvalue())
            self.assertEqual(retcode, 0)
        finally:
            os.close(fd)
            os.unlink(path)



# -*- coding: utf-8 -*-

# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import unittest2

from st2client.utils import strutil


class StrUtilTestCase(unittest2.TestCase):
    # See https://mail.python.org/pipermail/python-list/2006-January/411909.html

    def test_unescape(self):
        in_str = 'Action execution result double escape \\"stuffs\\".\\r\\n'
        expected = 'Action execution result double escape "stuffs".\r\n'
        out_str = strutil.unescape(in_str)
        self.assertEqual(out_str, expected)

    def test_unicode_string(self):
        in_str = "\u8c03\u7528CMS\u63a5\u53e3\u5220\u9664\u865a\u62df\u76ee\u5f55"
        out_str = strutil.unescape(in_str)
        self.assertEqual(out_str, in_str)

    def test_strip_carriage_returns(self):
        in_str = "Windows editors introduce\r\nlike a noob in 2017."
        out_str = strutil.strip_carriage_returns(in_str)
        exp_str = "Windows editors introduce\nlike a noob in 2017."
        self.assertEqual(out_str, exp_str)



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from st2client.models.reactor import RuleEnforcement
from st2client.commands import resource
from st2client.formatters import table
from st2client.utils.date import format_isodate_for_user_timezone


class RuleEnforcementBranch(resource.ResourceBranch):
    def __init__(self, description, app, subparsers, parent_parser=None):
        super(RuleEnforcementBranch, self).__init__(
            RuleEnforcement,
            description,
            app,
            subparsers,
            parent_parser=parent_parser,
            commands={
                "list": RuleEnforcementListCommand,
                "get": RuleEnforcementGetCommand,
            },
        )


class RuleEnforcementGetCommand(resource.ResourceGetCommand):
    display_attributes = [
        "id",
        "rule.ref",
        "trigger_instance_id",
        "execution_id",
        "failure_reason",
        "enforced_at",
    ]
    attribute_display_order = [
        "id",
        "rule.ref",
        "trigger_instance_id",
        "execution_id",
        "failure_reason",
        "enforced_at",
    ]

    pk_argument_name = "id"


class RuleEnforcementListCommand(resource.ResourceCommand):
    display_attributes = [
        "id",
        "rule.ref",
        "trigger_instance_id",
        "execution_id",
        "enforced_at",
    ]
    attribute_display_order = [
        "id",
        "rule.ref",
        "trigger_instance_id",
        "execution_id",
        "enforced_at",
    ]

    attribute_transform_functions = {"enforced_at": format_isodate_for_user_timezone}

    def __init__(self, resource, *args, **kwargs):

        self.default_limit = 50

        super(RuleEnforcementListCommand, self).__init__(
            resource,
            "list",
            "Get the list of the %s most recent %s."
            % (self.default_limit, resource.get_plural_display_name().lower()),
            *args,
            **kwargs,
        )
        self.resource_name = resource.get_plural_display_name().lower()
        self.group = self.parser.add_argument_group()
        self.parser.add_argument(
            "-n",
            "--last",
            type=int,
            dest="last",
            default=self.default_limit,
            help=(
                "List N most recent %s. Use -n -1 to fetch the full result \
                                       set."
                % self.resource_name
            ),
        )

        # Filter options
        self.group.add_argument(
            "--trigger-instance", help="Trigger instance id to filter the list."
        )

        self.group.add_argument("--execution", help="Execution id to filter the list.")
        self.group.add_argument("--rule", help="Rule ref to filter the list.")

        self.parser.add_argument(
            "-tg",
            "--timestamp-gt",
            type=str,
            dest="timestamp_gt",
            default=None,
            help=(
                "Only return enforcements with enforced_at "
                "greater than the one provided. "
                "Use time in the format 2000-01-01T12:00:00.000Z"
            ),
        )
        self.parser.add_argument(
            "-tl",
            "--timestamp-lt",
            type=str,
            dest="timestamp_lt",
            default=None,
            help=(
                "Only return enforcements with enforced_at "
                "lower than the one provided. "
                "Use time in the format 2000-01-01T12:00:00.000Z"
            ),
        )
        # Display options
        self.parser.add_argument(
            "-a",
            "--attr",
            nargs="+",
            default=self.display_attributes,
            help=(
                "List of attributes to include in the "
                'output. "all" will return all '
                "attributes."
            ),
        )
        self.parser.add_argument(
            "-w",
            "--width",
            nargs="+",
            type=int,
            default=None,
            help=("Set the width of columns in output."),
        )

    @resource.add_auth_token_to_kwargs_from_cli
    def run(self, args, **kwargs):
        # Filtering options
        if args.trigger_instance:
            kwargs["trigger_instance"] = args.trigger_instance
        if args.execution:
            kwargs["execution"] = args.execution
        if args.rule:
            kwargs["rule_ref"] = args.rule
        if args.timestamp_gt:
            kwargs["enforced_at_gt"] = args.timestamp_gt
        if args.timestamp_lt:
            kwargs["enforced_at_lt"] = args.timestamp_lt

        return self.manager.query_with_count(limit=args.last, **kwargs)

    def run_and_print(self, args, **kwargs):
        instances, count = self.run(args, **kwargs)
        if args.json or args.yaml:
            self.print_output(
                reversed(instances),
                table.MultiColumnTable,
                attributes=args.attr,
                widths=args.width,
                json=args.json,
                yaml=args.yaml,
                attribute_transform_functions=self.attribute_transform_functions,
            )
        else:
            self.print_output(
                instances,
                table.MultiColumnTable,
                attributes=args.attr,
                widths=args.width,
            )
            if args.last and count and count > args.last:
                table.SingleRowTable.note_box(self.resource_name, args.last)



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from typing import Optional
from typing import Tuple

import os
import logging
from functools import wraps

import six
import orjson
from six.moves import urllib
from six.moves import http_client
import requests

from st2client.utils import httpclient


LOG = logging.getLogger(__name__)


def add_auth_token_to_kwargs_from_env(func):
    @wraps(func)
    def decorate(*args, **kwargs):
        if not kwargs.get("token") and os.environ.get("ST2_AUTH_TOKEN", None):
            kwargs["token"] = os.environ.get("ST2_AUTH_TOKEN")
        if not kwargs.get("api_key") and os.environ.get("ST2_API_KEY", None):
            kwargs["api_key"] = os.environ.get("ST2_API_KEY")

        return func(*args, **kwargs)

    return decorate


def parse_api_response(response: requests.models.Response) -> dict:
    """
    Utility function which parses API json response using orjson.

    requests uses simplejson, but this version uses orjson which can be up to 50% faster and that
    is especially pronounced with larger response.
    """
    # Upstream implementation is available at
    # https://github.com/psf/requests/blob/master/requests/models.py#L876
    # NOTE: It's important that we manually decode response.content attribute before passing it to
    # orjson otherwise orjson will try to decode it and will be slower.

    # Inside tests content is not always bytes
    if isinstance(response.content, str):
        data = response.content
    else:
        data = response.content.decode("utf-8")

    data = orjson.loads(data)
    return data


class Resource(object):

    # An alias to use for the resource if different than the class name.
    _alias = None

    # Display name of the resource. This may be different than its resource
    # name specifically when the resource name is composed of multiple words.
    _display_name = None

    # URL path for the resource.
    _url_path = None

    # Plural form of the resource name. This will be used to build the
    # latter part of the REST URL.
    _plural = None

    # Plural form of the resource display name.
    _plural_display_name = None

    # A list of class attributes which will be included in __repr__ return value
    _repr_attributes = []

    def __init__(self, *args, **kwargs):
        for k, v in six.iteritems(kwargs):
            setattr(self, k, v)

    def to_dict(self, exclude_attributes=None):
        """
        Return a dictionary representation of this object.

        :param exclude_attributes: Optional list of attributes to exclude.
        :type exclude_attributes: ``list``

        :rtype: ``dict``
        """
        exclude_attributes = exclude_attributes or []

        attributes = list(self.__dict__.keys())
        attributes = [
            attr
            for attr in attributes
            if not attr.startswith("__") and attr not in exclude_attributes
        ]

        result = {}
        for attribute in attributes:
            value = getattr(self, attribute, None)
            result[attribute] = value

        return result

    @classmethod
    def get_alias(cls):
        return cls._alias if cls._alias else cls.__name__

    @classmethod
    def get_display_name(cls):
        return cls._display_name if cls._display_name else cls.__name__

    @classmethod
    def get_plural_name(cls):
        if not cls._plural:
            raise Exception(
                "The %s class is missing class attributes "
                "in its definition." % cls.__name__
            )
        return cls._plural

    @classmethod
    def get_plural_display_name(cls):
        return cls._plural_display_name if cls._plural_display_name else cls._plural

    @classmethod
    def get_url_path_name(cls):
        if cls._url_path:
            return cls._url_path

        return cls.get_plural_name().lower()

    def serialize(self):
        return dict(
            (k, v) for k, v in six.iteritems(self.__dict__) if not k.startswith("_")
        )

    @classmethod
    def deserialize(cls, doc):
        if type(doc) is not dict:
            doc = orjson.loads(doc)
        return cls(**doc)

    def __str__(self):
        return str(self.__repr__())

    def __repr__(self):
        if not self._repr_attributes:
            return super(Resource, self).__repr__()

        attributes = []
        for attribute in self._repr_attributes:
            value = getattr(self, attribute, None)
            attributes.append("%s=%s" % (attribute, value))

        attributes = ",".join(attributes)
        class_name = self.__class__.__name__
        result = "<%s %s>" % (class_name, attributes)
        return result


class ResourceManager(object):
    def __init__(
        self,
        resource: str,
        endpoint: str,
        cacert: Optional[str] = None,
        debug: bool = False,
        basic_auth: Optional[Tuple[str, str]] = None,
    ):
        """
        :param resource: Name of the resource to operate on.
        :param endpoint: API endpoint URL.
        :param cacert: Optional path to CA cert to use to validate the server side cert.
        :param debug: True to enable debug mode where additional debug information will be logged.
        :param basic_auth: Optional additional basic auth credentials in tuple(username, password)
                           notation.
        """
        self.resource = resource
        self.endpoint = endpoint
        self.cacert = cacert
        self.debug = debug
        self.basic_auth = basic_auth
        self.client = httpclient.HTTPClient(
            endpoint, cacert=cacert, debug=debug, basic_auth=basic_auth
        )

    @staticmethod
    def handle_error(response):
        try:
            content = parse_api_response(response)
            fault = content.get("faultstring", "") if content else ""
            if fault:
                response.reason += "\nMESSAGE: %s" % fault
        except Exception as e:
            response.reason += (
                "\nUnable to retrieve detailed message "
                "from the HTTP response. %s\n" % six.text_type(e)
            )
        response.raise_for_status()

    @add_auth_token_to_kwargs_from_env
    def get_all(self, **kwargs):
        # TODO: This is ugly, stop abusing kwargs
        url = "/%s" % self.resource.get_url_path_name()
        limit = kwargs.pop("limit", None)
        pack = kwargs.pop("pack", None)
        prefix = kwargs.pop("prefix", None)
        user = kwargs.pop("user", None)
        offset = kwargs.pop("offset", 0)

        params = kwargs.pop("params", {})

        if limit:
            params["limit"] = limit

        if pack:
            params["pack"] = pack

        if prefix:
            params["prefix"] = prefix

        if user:
            params["user"] = user

        if offset:
            params["offset"] = offset

        response = self.client.get(url=url, params=params, **kwargs)
        if response.status_code != http_client.OK:
            self.handle_error(response)
        return [
            self.resource.deserialize(item) for item in parse_api_response(response)
        ]

    @add_auth_token_to_kwargs_from_env
    def get_by_id(self, id, **kwargs):
        url = "/%s/%s" % (self.resource.get_url_path_name(), id)
        response = self.client.get(url, **kwargs)
        if response.status_code == http_client.NOT_FOUND:
            return None
        if response.status_code != http_client.OK:
            self.handle_error(response)

        return self.resource.deserialize(parse_api_response(response=response))

    @add_auth_token_to_kwargs_from_env
    def get_property(self, id_, property_name, self_deserialize=True, **kwargs):
        """
        Gets a property of a Resource.
        id_ : Id of the resource
        property_name: Name of the property
        self_deserialize: #Implies use the deserialize method implemented by this resource.
        """
        token = kwargs.pop("token", None)
        api_key = kwargs.pop("api_key", None)

        if kwargs:
            url = "/%s/%s/%s/?%s" % (
                self.resource.get_url_path_name(),
                id_,
                property_name,
                urllib.parse.urlencode(kwargs),
            )
        else:
            url = "/%s/%s/%s/" % (self.resource.get_url_path_name(), id_, property_name)

        if token:
            response = self.client.get(url, token=token)
        elif api_key:
            response = self.client.get(url, api_key=api_key)
        else:
            response = self.client.get(url)

        if response.status_code == http_client.NOT_FOUND:
            return None
        if response.status_code != http_client.OK:
            self.handle_error(response)

        if self_deserialize:
            return [
                self.resource.deserialize(item) for item in parse_api_response(response)
            ]
        else:
            return parse_api_response(response)

    @add_auth_token_to_kwargs_from_env
    def get_by_ref_or_id(self, ref_or_id, **kwargs):
        return self.get_by_id(id=ref_or_id, **kwargs)

    def _query_details(self, **kwargs):
        if not kwargs:
            raise Exception("Query parameter is not provided.")

        token = kwargs.get("token", None)
        api_key = kwargs.get("api_key", None)
        params = kwargs.get("params", {})

        for k, v in six.iteritems(kwargs):
            # Note: That's a special case to support api_key and token kwargs
            if k not in ["token", "api_key", "params"]:
                params[k] = v

        url = "/%s/?%s" % (
            self.resource.get_url_path_name(),
            urllib.parse.urlencode(params),
        )

        if token:
            response = self.client.get(url, token=token)
        elif api_key:
            response = self.client.get(url, api_key=api_key)
        else:
            response = self.client.get(url)

        if response.status_code == http_client.NOT_FOUND:
            # for query and query_with_count
            return [], None
        if response.status_code != http_client.OK:
            self.handle_error(response)
        items = parse_api_response(response)
        instances = [self.resource.deserialize(item) for item in items]
        return instances, response

    @add_auth_token_to_kwargs_from_env
    def query(self, **kwargs):
        instances, _ = self._query_details(**kwargs)
        return instances

    @add_auth_token_to_kwargs_from_env
    def query_with_count(self, **kwargs):
        instances, response = self._query_details(**kwargs)
        if response and "X-Total-Count" in response.headers:
            return (instances, int(response.headers["X-Total-Count"]))
        else:
            return (instances, None)

    @add_auth_token_to_kwargs_from_env
    def get_by_name(self, name, **kwargs):
        instances = self.query(name=name, **kwargs)
        if not instances:
            return None
        else:
            if len(instances) > 1:
                raise Exception(
                    'More than one %s named "%s" are found.'
                    % (self.resource.__name__.lower(), name)
                )
            return instances[0]

    @add_auth_token_to_kwargs_from_env
    def create(self, instance, **kwargs):
        url = "/%s" % self.resource.get_url_path_name()
        response = self.client.post(url, instance.serialize(), **kwargs)
        if response.status_code != http_client.OK:
            self.handle_error(response)
        instance = self.resource.deserialize(parse_api_response(response))
        return instance

    @add_auth_token_to_kwargs_from_env
    def update(self, instance, **kwargs):
        url = "/%s/%s" % (self.resource.get_url_path_name(), instance.id)
        response = self.client.put(url, instance.serialize(), **kwargs)
        if response.status_code != http_client.OK:
            self.handle_error(response)
        instance = self.resource.deserialize(parse_api_response(response))
        return instance

    @add_auth_token_to_kwargs_from_env
    def delete(self, instance, **kwargs):
        url = "/%s/%s" % (self.resource.get_url_path_name(), instance.id)
        response = self.client.delete(url, **kwargs)

        if response.status_code not in [
            http_client.OK,
            http_client.NO_CONTENT,
            http_client.NOT_FOUND,
        ]:
            self.handle_error(response)
            return False

        return True

    @add_auth_token_to_kwargs_from_env
    def delete_action(self, instance, remove_files, **kwargs):
        url = "/%s/%s" % (self.resource.get_url_path_name(), instance.id)
        payload = {"remove_files": remove_files}
        response = self.client.delete(url, data=orjson.dumps(payload), **kwargs)
        if response.status_code not in [
            http_client.OK,
            http_client.NO_CONTENT,
            http_client.NOT_FOUND,
        ]:
            self.handle_error(response)
            return False

        return True

    @add_auth_token_to_kwargs_from_env
    def clone(
        self,
        source_ref,
        dest_pack,
        dest_action,
        overwrite,
        **kwargs,
    ):
        url = "/%s/%s/clone" % (self.resource.get_url_path_name(), source_ref)
        payload = {
            "dest_pack": dest_pack,
            "dest_action": dest_action,
            "overwrite": overwrite,
        }
        response = self.client.post(url, payload, **kwargs)
        if response.status_code != http_client.OK:
            self.handle_error(response)
        instance = self.resource.deserialize(parse_api_response(response))
        return instance

    @add_auth_token_to_kwargs_from_env
    def delete_by_id(self, instance_id, **kwargs):
        url = "/%s/%s" % (self.resource.get_url_path_name(), instance_id)
        response = self.client.delete(url, **kwargs)
        if response.status_code not in [
            http_client.OK,
            http_client.NO_CONTENT,
            http_client.NOT_FOUND,
        ]:
            self.handle_error(response)
            return False
        try:
            resp_json = parse_api_response(response)
            if resp_json:
                return resp_json
        except Exception as e:
            print(
                "\nUnable to retrieve detailed message "
                "from the HTTP response. %s\n" % six.text_type(e)
            )
        return True


class ActionAliasResourceManager(ResourceManager):
    @add_auth_token_to_kwargs_from_env
    def match(self, instance, **kwargs):
        url = "/%s/match" % self.resource.get_url_path_name()
        response = self.client.post(url, instance.serialize(), **kwargs)
        if response.status_code != http_client.OK:
            self.handle_error(response)
        match = parse_api_response(response)
        return (
            self.resource.deserialize(match["actionalias"]),
            match["representation"],
        )


class ActionAliasExecutionManager(ResourceManager):
    @add_auth_token_to_kwargs_from_env
    def match_and_execute(self, instance, **kwargs):
        url = "/%s/match_and_execute" % self.resource.get_url_path_name()
        response = self.client.post(url, instance.serialize(), **kwargs)

        if response.status_code != http_client.OK:
            self.handle_error(response)
        instance = self.resource.deserialize(parse_api_response(response)["results"][0])
        return instance


class ActionResourceManager(ResourceManager):
    @add_auth_token_to_kwargs_from_env
    def get_entrypoint(self, ref_or_id, **kwargs):
        url = "/%s/views/entry_point/%s" % (
            self.resource.get_url_path_name(),
            ref_or_id,
        )

        response = self.client.get(url, **kwargs)
        if response.status_code != http_client.OK:
            self.handle_error(response)

        return response.text


class ExecutionResourceManager(ResourceManager):
    @add_auth_token_to_kwargs_from_env
    def re_run(
        self,
        execution_id,
        parameters=None,
        tasks=None,
        no_reset=None,
        delay=0,
        **kwargs,
    ):
        url = "/%s/%s/re_run" % (self.resource.get_url_path_name(), execution_id)

        tasks = tasks or []
        no_reset = no_reset or []

        if list(set(no_reset) - set(tasks)):
            raise ValueError(
                "List of tasks to reset does not match the tasks to rerun."
            )

        data = {
            "parameters": parameters or {},
            "tasks": tasks,
            "reset": list(set(tasks) - set(no_reset)),
            "delay": delay,
        }

        response = self.client.post(url, data, **kwargs)
        if response.status_code != http_client.OK:
            self.handle_error(response)

        instance = self.resource.deserialize(parse_api_response(response))
        return instance

    @add_auth_token_to_kwargs_from_env
    def get_output(self, execution_id, output_type=None, **kwargs):
        url = "/%s/%s/output" % (self.resource.get_url_path_name(), execution_id)

        if output_type:
            url += "?" + urllib.parse.urlencode({"output_type": output_type})

        response = self.client.get(url, **kwargs)
        if response.status_code != http_client.OK:
            self.handle_error(response)

        return response.text

    @add_auth_token_to_kwargs_from_env
    def pause(self, execution_id, **kwargs):
        url = "/%s/%s" % (self.resource.get_url_path_name(), execution_id)
        data = {"status": "pausing"}

        response = self.client.put(url, data, **kwargs)

        if response.status_code != http_client.OK:
            self.handle_error(response)

        return self.resource.deserialize(parse_api_response(response))

    @add_auth_token_to_kwargs_from_env
    def resume(self, execution_id, **kwargs):
        url = "/%s/%s" % (self.resource.get_url_path_name(), execution_id)
        data = {"status": "resuming"}

        response = self.client.put(url, data, **kwargs)

        if response.status_code != http_client.OK:
            self.handle_error(response)

        return self.resource.deserialize(parse_api_response(response))

    @add_auth_token_to_kwargs_from_env
    def get_children(self, execution_id, **kwargs):
        url = "/%s/%s/children" % (self.resource.get_url_path_name(), execution_id)

        depth = kwargs.pop("depth", -1)

        params = kwargs.pop("params", {})

        if depth:
            params["depth"] = depth

        response = self.client.get(url=url, params=params, **kwargs)
        if response.status_code != http_client.OK:
            self.handle_error(response)
        return [
            self.resource.deserialize(item) for item in parse_api_response(response)
        ]


class InquiryResourceManager(ResourceManager):
    @add_auth_token_to_kwargs_from_env
    def respond(self, inquiry_id, inquiry_response, **kwargs):
        """
        Update st2.inquiry.respond action
        Update st2client respond command to use this?
        """
        url = "/%s/%s" % (self.resource.get_url_path_name(), inquiry_id)

        payload = {"id": inquiry_id, "response": inquiry_response}

        response = self.client.put(url, payload, **kwargs)

        if response.status_code != http_client.OK:
            self.handle_error(response)

        return self.resource.deserialize(parse_api_response(response))


class TriggerInstanceResourceManager(ResourceManager):
    @add_auth_token_to_kwargs_from_env
    def re_emit(self, trigger_instance_id, **kwargs):
        url = "/%s/%s/re_emit" % (
            self.resource.get_url_path_name(),
            trigger_instance_id,
        )
        response = self.client.post(url, None, **kwargs)
        if response.status_code != http_client.OK:
            self.handle_error(response)
        return parse_api_response(response)


class AsyncRequest(Resource):
    pass


class PackResourceManager(ResourceManager):
    @add_auth_token_to_kwargs_from_env
    def install(self, packs, force=False, skip_dependencies=False, **kwargs):
        url = "/%s/install" % (self.resource.get_url_path_name())
        payload = {
            "packs": packs,
            "force": force,
            "skip_dependencies": skip_dependencies,
        }
        response = self.client.post(url, payload, **kwargs)
        if response.status_code != http_client.OK:
            self.handle_error(response)
        instance = AsyncRequest.deserialize(parse_api_response(response))
        return instance

    @add_auth_token_to_kwargs_from_env
    def remove(self, packs, **kwargs):
        url = "/%s/uninstall" % (self.resource.get_url_path_name())
        response = self.client.post(url, {"packs": packs}, **kwargs)
        if response.status_code != http_client.OK:
            self.handle_error(response)
        instance = AsyncRequest.deserialize(parse_api_response(response))
        return instance

    @add_auth_token_to_kwargs_from_env
    def search(self, args, ignore_errors=False, **kwargs):
        url = "/%s/index/search" % (self.resource.get_url_path_name())
        if "query" in vars(args):
            payload = {"query": args.query}
        else:
            payload = {"pack": args.pack}

        response = self.client.post(url, payload, **kwargs)

        if response.status_code != http_client.OK:
            if ignore_errors:
                return None

            self.handle_error(response)
        data = parse_api_response(response)
        if isinstance(data, list):
            return [self.resource.deserialize(item) for item in data]
        else:
            return self.resource.deserialize(data) if data else None

    @add_auth_token_to_kwargs_from_env
    def register(self, packs=None, types=None, **kwargs):
        url = "/%s/register" % (self.resource.get_url_path_name())
        payload = {}
        if types:
            payload["types"] = types
        if packs:
            payload["packs"] = packs
        response = self.client.post(url, payload, **kwargs)
        if response.status_code != http_client.OK:
            self.handle_error(response)
        instance = self.resource.deserialize(parse_api_response(response))
        return instance


class ConfigManager(ResourceManager):
    @add_auth_token_to_kwargs_from_env
    def update(self, instance, **kwargs):
        url = "/%s/%s" % (self.resource.get_url_path_name(), instance.pack)
        response = self.client.put(url, instance.values, **kwargs)
        if response.status_code != http_client.OK:
            self.handle_error(response)
        instance = self.resource.deserialize(parse_api_response(response))
        return instance


class WebhookManager(ResourceManager):
    @add_auth_token_to_kwargs_from_env
    def post_generic_webhook(self, trigger, payload=None, trace_tag=None, **kwargs):
        url = "/webhooks/st2"

        headers = {}
        data = {"trigger": trigger, "payload": payload or {}}

        if trace_tag:
            headers["St2-Trace-Tag"] = trace_tag

        response = self.client.post(url, data=data, headers=headers, **kwargs)

        if response.status_code != http_client.OK:
            self.handle_error(response)

        return parse_api_response(response)

    @add_auth_token_to_kwargs_from_env
    def match(self, instance, **kwargs):
        url = "/%s/match" % self.resource.get_url_path_name()
        response = self.client.post(url, instance.serialize(), **kwargs)
        if response.status_code != http_client.OK:
            self.handle_error(response)
        match = parse_api_response(response)
        return (
            self.resource.deserialize(match["actionalias"]),
            match["representation"],
        )


class StreamManager(ResourceManager):
    def __init__(self, endpoint, cacert=None, debug=False, basic_auth=None):
        super(StreamManager, self).__init__(
            resource=None,
            endpoint=endpoint,
            cacert=cacert,
            debug=debug,
            basic_auth=basic_auth,
        )
        self._url = httpclient.get_url_without_trailing_slash(endpoint) + "/stream"

    @add_auth_token_to_kwargs_from_env
    def listen(self, events=None, **kwargs):
        # Late import to avoid very expensive in-direct import (~1 second) when this function is
        # not called / used
        from sseclient import SSEClient

        url = self._url
        query_params = {}
        request_params = {}

        if events and isinstance(events, six.string_types):
            events = [events]

        if "token" in kwargs:
            query_params["x-auth-token"] = kwargs.get("token")

        if "api_key" in kwargs:
            query_params["st2-api-key"] = kwargs.get("api_key")

        if "end_event" in kwargs:
            query_params["end_event"] = kwargs.get("end_event")

        if "end_execution_id" in kwargs:
            query_params["end_execution_id"] = kwargs.get("end_execution_id")

        if events:
            query_params["events"] = ",".join(events)

        if self.cacert is not None:
            request_params["verify"] = self.cacert

        if self.basic_auth:
            request_params["auth"] = self.basic_auth

        query_string = "?" + urllib.parse.urlencode(query_params)
        url = url + query_string

        response = requests.get(url, stream=True, **request_params)
        client = SSEClient(response)

        for message in client.events():
            # If the execution on the API server takes too long, the message
            # can be empty. In this case, rerun the query.
            if not message.data:
                continue
            yield orjson.loads(message.data)


class WorkflowManager(object):
    def __init__(self, endpoint, cacert, debug, basic_auth=None):
        self.debug = debug
        self.cacert = cacert
        self.endpoint = endpoint + "/workflows"
        self.client = httpclient.HTTPClient(
            root=self.endpoint, cacert=cacert, debug=debug, basic_auth=basic_auth
        )

    @staticmethod
    def handle_error(response):
        try:
            content = parse_api_response(response)
            fault = content.get("faultstring", "") if content else ""

            if fault:
                response.reason += "\nMESSAGE: %s" % fault
        except Exception as e:
            response.reason += (
                "\nUnable to retrieve detailed message "
                "from the HTTP response. %s\n" % six.text_type(e)
            )

        response.raise_for_status()

    @add_auth_token_to_kwargs_from_env
    def inspect(self, definition, **kwargs):
        url = "/inspect"

        if not isinstance(definition, six.string_types):
            raise TypeError(
                f"Workflow definition is not type of string (was {type(definition)})."
            )

        if "headers" not in kwargs:
            kwargs["headers"] = {}

        kwargs["headers"]["content-type"] = "text/plain"

        response = self.client.post_raw(url, definition, **kwargs)

        if response.status_code != http_client.OK:
            self.handle_error(response)

        return parse_api_response(response)


class ServiceRegistryGroupsManager(ResourceManager):
    @add_auth_token_to_kwargs_from_env
    def list(self, **kwargs):
        url = "/service_registry/groups"

        headers = {}
        response = self.client.get(url, headers=headers, **kwargs)

        if response.status_code != http_client.OK:
            self.handle_error(response)

        groups = parse_api_response(response)["groups"]

        result = []
        for group in groups:
            item = self.resource.deserialize({"group_id": group})
            result.append(item)

        return result


class ServiceRegistryMembersManager(ResourceManager):
    @add_auth_token_to_kwargs_from_env
    def list(self, group_id, **kwargs):
        url = "/service_registry/groups/%s/members" % (group_id)

        headers = {}
        response = self.client.get(url, headers=headers, **kwargs)

        if response.status_code != http_client.OK:
            self.handle_error(response)

        members = parse_api_response(response)["members"]

        result = []
        for member in members:
            data = {
                "group_id": group_id,
                "member_id": member["member_id"],
                "capabilities": member["capabilities"],
            }
            item = self.resource.deserialize(data)
            result.append(item)

        return result


class KeyValuePairResourceManager(ResourceManager):
    @add_auth_token_to_kwargs_from_env
    def get_by_name(self, name, **kwargs):

        token = kwargs.get("token", None)
        api_key = kwargs.get("api_key", None)
        params = kwargs.get("params", {})

        for k, v in six.iteritems(kwargs):
            # Note: That's a special case to support api_key and token kwargs
            if k not in ["token", "api_key", "params"]:
                params[k] = v

        url = "/%s/%s/?%s" % (
            self.resource.get_url_path_name(),
            name,
            urllib.parse.urlencode(params),
        )

        if token:
            response = self.client.get(url, token=token)
        elif api_key:
            response = self.client.get(url, api_key=api_key)
        else:
            response = self.client.get(url)

        if response.status_code == http_client.NOT_FOUND:
            # for query and query_with_count
            return []
        if response.status_code != http_client.OK:
            self.handle_error(response)

        return self.resource.deserialize(parse_api_response(response))



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import ast
import logging
import struct

import yaml

try:
    from yaml import CSafeDumper as YamlSafeDumper
except ImportError:
    # NOTE: We install libyaml-dev in our packages so libyaml will always be available when using
    # official StackStorm packages.
    # Only time it may not be available is if the user is doing custom install from source or
    # similar.
    logging.getLogger(__name__).warn(
        "libYAML C bindings are not available. This means YAML "
        "parsing and serialization will be significantly slower. You are "
        "strongly recommended to install libyaml (libyaml-dev package "
        "on Debian). For more information, see https://pyyaml.org/wiki/LibYAML"
    )
    from yaml import SafeDumper as YamlSafeDumper

from st2client import formatters
from st2client.config import get_config
from st2client.utils import jsutil
from st2client.utils import strutil
from st2client.utils.color import DisplayColors
from st2client.utils import schema
import six


LOG = logging.getLogger(__name__)

PLATFORM_MAXINT = 2 ** (struct.Struct("i").size * 8 - 1) - 1


def _print_bordered(text):
    lines = text.split("\n")
    width = max(len(s) for s in lines) + 2
    res = ["\n+" + "-" * width + "+"]
    for s in lines:
        res.append("| " + (s + " " * width)[: width - 2] + " |")
    res.append("+" + "-" * width + "+")
    return "\n".join(res)


class ExecutionResult(formatters.Formatter):
    @classmethod
    def format(cls, entry, *args, **kwargs):
        attrs = kwargs.get("attributes", [])
        attribute_transform_functions = kwargs.get("attribute_transform_functions", {})
        key = kwargs.get("key", None)
        if key:
            output = jsutil.get_value(entry.result, key)
        else:
            # drop entry to the dict so that jsutil can operate
            entry = vars(entry)
            output = ""
            for attr in attrs:
                value = jsutil.get_value(entry, attr)
                value = strutil.strip_carriage_returns(strutil.unescape(value))

                # transform the value of our attribute so things like 'status'
                # and 'timestamp' are formatted nicely
                transform_function = attribute_transform_functions.get(
                    attr, lambda value: value
                )
                value = transform_function(value=value)

                # TODO: This check is inherently flawed since it will crash st2client
                # if the leading character is objectish start and last character is objectish
                # end but the string isn't supposed to be a object. Try/Except will catch
                # this for now, but this should be improved.
                if (
                    isinstance(value, six.string_types)
                    and len(value) > 0
                    and value[0] in ["{", "["]
                    and value[len(value) - 1] in ["}", "]"]
                ):
                    try:
                        new_value = ast.literal_eval(value)
                    except:
                        new_value = value
                    if type(new_value) in [dict, list]:
                        value = new_value
                if isinstance(value, (dict, list)):
                    # 1. To get a nice overhang indent get safe_dump to generate output with
                    #    the attribute key and then remove the attribute key from the string.
                    # 2. Drop the trailing newline
                    # 3. Set width to maxint so pyyaml does not split text. Anything longer
                    #    and likely we will see other issues like storage :P.
                    # NOTE: We use C version of the safe dumper which is faster.
                    # Keep in mind that using YamlSafeDumper is the same as using yaml.safe_dumps
                    # (same class is used underneath when using yaml.safe_dump) so the code is safe.
                    formatted_value = yaml.dump(
                        {attr: value},
                        default_flow_style=False,
                        width=PLATFORM_MAXINT,
                        indent=2,
                        Dumper=YamlSafeDumper,
                    )[len(attr) + 2 : -1]

                    if isinstance(value, list):
                        # Indent list values with 2 spaces for a nicer display.
                        lines = formatted_value.split("\n")
                        formatted_value = []
                        for line in lines:
                            formatted_value.append("  %s" % (line))

                        formatted_value = "\n".join(formatted_value)

                    value = (
                        "\n" if isinstance(value, (dict, list)) else ""
                    ) + formatted_value
                    value = strutil.dedupe_newlines(value)

                # transform the value of our attribute so things like 'status'
                # and 'timestamp' are formatted nicely
                transform_function = attribute_transform_functions.get(
                    attr, lambda value: value
                )
                value = transform_function(value=value)

                output += ("\n" if output else "") + "%s: %s" % (
                    DisplayColors.colorize(attr, DisplayColors.BLUE),
                    value,
                )

            output_schema = entry.get("action", {}).get("output_schema")
            schema_check = get_config()["general"]["silence_schema_output"]
            if not output_schema and kwargs.get("with_schema"):
                rendered_schema = {
                    "output_schema": schema.render_output_schema_from_output(
                        entry["result"]
                    )
                }

                rendered_schema = yaml.safe_dump(
                    rendered_schema, default_flow_style=False
                )
                output += "\n"
                output += _print_bordered(
                    "Based on the action output the following inferred schema was built:"
                    "\n\n"
                    "%s" % rendered_schema
                )
            elif not output_schema and not schema_check:
                output += (
                    "\n\n** This action does not have an output_schema. "
                    "Run again with --with-schema to see a suggested schema."
                )

        if six.PY3:
            return strutil.unescape(str(output))
        else:
            # Assume Python 2
            try:
                result = (
                    strutil.unescape(str(output))
                    .decode("unicode_escape")
                    .encode("utf-8")
                )
            except UnicodeDecodeError:
                # String contains a value which is not an unicode escape sequence, ignore the error
                result = strutil.unescape(str(output))
            return result



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from st2client.commands import resource
from st2client.models.reactor import TriggerType
from st2client.formatters import table


class TriggerTypeBranch(resource.ResourceBranch):
    def __init__(self, description, app, subparsers, parent_parser=None):
        super(TriggerTypeBranch, self).__init__(
            TriggerType,
            description,
            app,
            subparsers,
            parent_parser=parent_parser,
            commands={
                "list": TriggerTypeListCommand,
                "get": TriggerTypeGetCommand,
                "update": TriggerTypeUpdateCommand,
                "delete": TriggerTypeDeleteCommand,
            },
        )

        # Registers extended commands
        self.commands["getspecs"] = TriggerTypeSubTriggerCommand(
            self.resource, self.app, self.subparsers, add_help=False
        )


class TriggerTypeListCommand(resource.ContentPackResourceListCommand):
    display_attributes = ["ref", "pack", "description"]


class TriggerTypeGetCommand(resource.ContentPackResourceGetCommand):
    display_attributes = ["all"]
    attribute_display_order = [
        "id",
        "ref",
        "pack",
        "name",
        "description",
        "parameters_schema",
        "payload_schema",
    ]


class TriggerTypeUpdateCommand(resource.ContentPackResourceUpdateCommand):
    pass


class TriggerTypeDeleteCommand(resource.ContentPackResourceDeleteCommand):
    pass


class TriggerTypeSubTriggerCommand(resource.ResourceCommand):
    attribute_display_order = [
        "id",
        "ref",
        "context",
        "parameters",
        "status",
        "start_timestamp",
        "result",
    ]

    def __init__(self, resource, *args, **kwargs):

        super(TriggerTypeSubTriggerCommand, self).__init__(
            resource,
            kwargs.pop("name", "getspecs"),
            "Return Trigger Specifications of a Trigger.",
            *args,
            **kwargs,
        )

        self.parser.add_argument(
            "ref",
            nargs="?",
            metavar="ref",
            help="Fully qualified name (pack.trigger_name) " + "of the trigger.",
        )

        self.parser.add_argument(
            "-h",
            "--help",
            action="store_true",
            dest="help",
            help="Print usage for the given action.",
        )

    @resource.add_auth_token_to_kwargs_from_cli
    def run(self, args, **kwargs):
        trigger_mgr = self.app.client.managers["Trigger"]
        return trigger_mgr.query(**{"type": args.ref})

    @resource.add_auth_token_to_kwargs_from_cli
    def run_and_print(self, args, **kwargs):
        if args.help:
            self.parser.print_help()
            return
        instances = self.run(args, **kwargs)
        self.print_output(
            instances, table.MultiColumnTable, json=args.json, yaml=args.yaml
        )



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import json
import logging
import mock
import unittest2

from tests import base

from st2client import client
from st2client.utils import httpclient


LOG = logging.getLogger(__name__)


EXECUTION = {
    "id": 12345,
    "action": {"ref": "mock.foobar"},
    "status": "failed",
    "result": "non-empty",
}

ENTRYPOINT = (
    "version: 1.0"
    "description: A basic workflow that runs an arbitrary linux command."
    "input:"
    "  - cmd"
    "  - timeout"
    "tasks:"
    "  task1:"
    "    action: core.local cmd=<% ctx(cmd) %> timeout=<% ctx(timeout) %>"
    "    next:"
    "      - when: <% succeeded() %>"
    "        publish:"
    "          - stdout: <% result().stdout %>"
    "          - stderr: <% result().stderr %>"
    "output:"
    "  - stdout: <% ctx(stdout) %>"
)


class TestActionResourceManager(unittest2.TestCase):
    @classmethod
    def setUpClass(cls):
        super(TestActionResourceManager, cls).setUpClass()
        cls.client = client.Client()

    @mock.patch.object(
        httpclient.HTTPClient,
        "get",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(ENTRYPOINT), 200, "OK")
        ),
    )
    def test_get_action_entry_point_by_ref(self):
        actual_entrypoint = self.client.actions.get_entrypoint(
            EXECUTION["action"]["ref"]
        )
        actual_entrypoint = json.loads(actual_entrypoint)

        endpoint = "/actions/views/entry_point/%s" % EXECUTION["action"]["ref"]
        httpclient.HTTPClient.get.assert_called_with(endpoint)

        self.assertEqual(ENTRYPOINT, actual_entrypoint)

    @mock.patch.object(
        httpclient.HTTPClient,
        "get",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(ENTRYPOINT), 200, "OK")
        ),
    )
    def test_get_action_entry_point_by_id(self):
        actual_entrypoint = self.client.actions.get_entrypoint(EXECUTION["id"])
        actual_entrypoint = json.loads(actual_entrypoint)

        endpoint = "/actions/views/entry_point/%s" % EXECUTION["id"]
        httpclient.HTTPClient.get.assert_called_with(endpoint)

        self.assertEqual(ENTRYPOINT, actual_entrypoint)

    @mock.patch.object(
        httpclient.HTTPClient,
        "get",
        mock.MagicMock(
            return_value=base.FakeResponse(
                json.dumps({}), 404, "404 Client Error: Not Found"
            )
        ),
    )
    def test_get_non_existent_action_entry_point(self):
        with self.assertRaisesRegexp(Exception, "404 Client Error: Not Found"):
            self.client.actions.get_entrypoint("nonexistentpack.nonexistentaction")

        endpoint = "/actions/views/entry_point/%s" % "nonexistentpack.nonexistentaction"
        httpclient.HTTPClient.get.assert_called_with(endpoint)



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import datetime

import mock
import unittest2

from st2client.utils.date import add_utc_tz
from st2client.utils.date import format_dt
from st2client.utils.date import format_isodate
from st2client.utils.date import format_isodate_for_user_timezone


class DateUtilsTestCase(unittest2.TestCase):
    def test_format_dt(self):
        dt = datetime.datetime(2015, 10, 20, 8, 0, 0)
        dt = add_utc_tz(dt)
        result = format_dt(dt)
        self.assertEqual(result, "Tue, 20 Oct 2015 08:00:00 UTC")

    def test_format_isodate(self):
        # No timezone, defaults to UTC
        value = "Tue, 20 Oct 2015 08:00:00 UTC"
        result = format_isodate(value=value)
        self.assertEqual(result, "Tue, 20 Oct 2015 08:00:00 UTC")

        # Timezone provided
        value = "Tue, 20 Oct 2015 08:00:00 UTC"
        result = format_isodate(value=value, timezone="Europe/Ljubljana")
        self.assertEqual(result, "Tue, 20 Oct 2015 10:00:00 CEST")

    @mock.patch("st2client.utils.date.get_config")
    def test_format_isodate_for_user_timezone(self, mock_get_config):
        # No timezone, defaults to UTC
        mock_get_config.return_value = {}

        value = "Tue, 20 Oct 2015 08:00:00 UTC"
        result = format_isodate_for_user_timezone(value=value)
        self.assertEqual(result, "Tue, 20 Oct 2015 08:00:00 UTC")

        # Timezone provided
        mock_get_config.return_value = {"cli": {"timezone": "Europe/Ljubljana"}}

        value = "Tue, 20 Oct 2015 08:00:00 UTC"
        result = format_isodate_for_user_timezone(value=value)
        self.assertEqual(result, "Tue, 20 Oct 2015 10:00:00 CEST")



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import logging

from st2client.models import core


LOG = logging.getLogger(__name__)


class Sensor(core.Resource):
    _plural = "Sensortypes"
    _repr_attributes = ["name", "pack"]


class TriggerType(core.Resource):
    _alias = "Trigger"
    _display_name = "Trigger"
    _plural = "Triggertypes"
    _plural_display_name = "Triggers"
    _repr_attributes = ["name", "pack"]


class TriggerInstance(core.Resource):
    _alias = "Trigger-Instance"
    _display_name = "Trigger Instance"
    _plural = "Triggerinstances"
    _plural_display_name = "TriggerInstances"
    _repr_attributes = ["id", "trigger", "occurrence_time", "payload", "status"]


class Trigger(core.Resource):
    _alias = "TriggerSpecification"
    _display_name = "Trigger Specification"
    _plural = "Triggers"
    _plural_display_name = "Trigger Specifications"
    _repr_attributes = ["name", "pack"]


class Rule(core.Resource):
    _alias = "Rule"
    _plural = "Rules"
    _repr_attributes = ["name", "pack", "trigger", "criteria", "enabled"]


class RuleEnforcement(core.Resource):
    _alias = "Rule-Enforcement"
    _plural = "RuleEnforcements"
    _display_name = "Rule Enforcement"
    _plural_display_name = "Rule Enforcements"
    _repr_attributes = [
        "id",
        "trigger_instance_id",
        "execution_id",
        "rule.ref",
        "enforced_at",
    ]



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import os
import json
import logging
import unittest2

import six
import mock
import requests

from st2client import models
from st2client.client import Client

from tests import base


LOG = logging.getLogger(__name__)

NONRESOURCES = ["workflows"]


class TestClientEndpoints(unittest2.TestCase):
    def tearDown(self):
        for var in [
            "ST2_BASE_URL",
            "ST2_API_URL",
            "ST2_STREAM_URL",
            "ST2_DATASTORE_URL",
            "ST2_AUTH_TOKEN",
        ]:
            if var in os.environ:
                del os.environ[var]

    def test_managers(self):
        property_names = [
            k for k, v in six.iteritems(Client.__dict__) if isinstance(v, property)
        ]

        client = Client()

        for property_name in property_names:
            manager = getattr(client, property_name, None)
            self.assertIsNotNone(manager)

            if property_name not in NONRESOURCES:
                self.assertIsInstance(manager, models.ResourceManager)

    def test_default(self):
        base_url = "http://127.0.0.1"
        api_url = "http://127.0.0.1:9101/v1"
        stream_url = "http://127.0.0.1:9102/v1"

        client = Client()
        endpoints = client.endpoints
        self.assertEqual(endpoints["base"], base_url)
        self.assertEqual(endpoints["api"], api_url)
        self.assertEqual(endpoints["stream"], stream_url)

    @mock.patch.object(
        requests,
        "get",
        mock.MagicMock(return_value=base.FakeResponse(json.dumps({}), 200, "OK")),
    )
    def test_basic_auth_option_success(self):
        client = Client(basic_auth="username:password")
        self.assertEqual(client.basic_auth, ("username", "password"))

        self.assertEqual(requests.get.call_count, 0)
        client.actions.get_all()
        self.assertEqual(requests.get.call_count, 1)

        requests.get.assert_called_with(
            "http://127.0.0.1:9101/v1/actions", auth=("username", "password"), params={}
        )

    @mock.patch.object(
        requests,
        "get",
        mock.MagicMock(return_value=base.FakeResponse(json.dumps({}), 200, "OK")),
    )
    def test_basic_auth_option_success_password_with_colon(self):
        client = Client(basic_auth="username:password:with:colon")
        self.assertEqual(client.basic_auth, ("username", "password:with:colon"))

        self.assertEqual(requests.get.call_count, 0)
        client.actions.get_all()
        self.assertEqual(requests.get.call_count, 1)

        requests.get.assert_called_with(
            "http://127.0.0.1:9101/v1/actions",
            auth=("username", "password:with:colon"),
            params={},
        )

    def test_basic_auth_option_invalid_notation(self):
        self.assertRaisesRegex(
            ValueError,
            "needs to be in the username:password notation",
            Client,
            basic_auth="username_password",
        )

    def test_env(self):
        base_url = "http://www.stackstorm.com"
        api_url = "http://www.st2.com:9101/v1"
        stream_url = "http://www.st2.com:9102/v1"

        os.environ["ST2_BASE_URL"] = base_url
        os.environ["ST2_API_URL"] = api_url
        os.environ["ST2_STREAM_URL"] = stream_url
        self.assertEqual(os.environ.get("ST2_BASE_URL"), base_url)
        self.assertEqual(os.environ.get("ST2_API_URL"), api_url)
        self.assertEqual(os.environ.get("ST2_STREAM_URL"), stream_url)

        client = Client()
        endpoints = client.endpoints
        self.assertEqual(endpoints["base"], base_url)
        self.assertEqual(endpoints["api"], api_url)
        self.assertEqual(endpoints["stream"], stream_url)

    def test_env_base_only(self):
        base_url = "http://www.stackstorm.com"
        api_url = "http://www.stackstorm.com:9101/v1"
        stream_url = "http://www.stackstorm.com:9102/v1"

        os.environ["ST2_BASE_URL"] = base_url
        self.assertEqual(os.environ.get("ST2_BASE_URL"), base_url)
        self.assertEqual(os.environ.get("ST2_API_URL"), None)
        self.assertEqual(os.environ.get("ST2_STREAM_URL"), None)

        client = Client()
        endpoints = client.endpoints
        self.assertEqual(endpoints["base"], base_url)
        self.assertEqual(endpoints["api"], api_url)
        self.assertEqual(endpoints["stream"], stream_url)

    def test_args(self):
        base_url = "http://www.stackstorm.com"
        api_url = "http://www.st2.com:9101/v1"
        stream_url = "http://www.st2.com:9102/v1"

        client = Client(base_url=base_url, api_url=api_url, stream_url=stream_url)
        endpoints = client.endpoints
        self.assertEqual(endpoints["base"], base_url)
        self.assertEqual(endpoints["api"], api_url)
        self.assertEqual(endpoints["stream"], stream_url)

    def test_cacert_arg(self):
        # Valid value, boolean True
        base_url = "http://www.stackstorm.com"
        api_url = "http://www.st2.com:9101/v1"
        stream_url = "http://www.st2.com:9102/v1"

        client = Client(
            base_url=base_url, api_url=api_url, stream_url=stream_url, cacert=True
        )
        self.assertEqual(client.cacert, True)

        # Valid value, boolean False
        base_url = "http://www.stackstorm.com"
        api_url = "http://www.st2.com:9101/v1"
        stream_url = "http://www.st2.com:9102/v1"

        client = Client(
            base_url=base_url, api_url=api_url, stream_url=stream_url, cacert=False
        )
        self.assertEqual(client.cacert, False)

        # Valid value, existing path to a CA bundle
        cacert = os.path.abspath(__file__)
        client = Client(
            base_url=base_url, api_url=api_url, stream_url=stream_url, cacert=cacert
        )
        self.assertEqual(client.cacert, cacert)

        # Invalid value, path to the bundle doesn't exist
        cacert = os.path.abspath(__file__)
        expected_msg = 'CA cert file "doesntexist" does not exist'
        self.assertRaisesRegexp(
            ValueError,
            expected_msg,
            Client,
            base_url=base_url,
            api_url=api_url,
            stream_url=stream_url,
            cacert="doesntexist",
        )

    def test_args_base_only(self):
        base_url = "http://www.stackstorm.com"
        api_url = "http://www.stackstorm.com:9101/v1"
        stream_url = "http://www.stackstorm.com:9102/v1"

        client = Client(base_url=base_url)
        endpoints = client.endpoints
        self.assertEqual(endpoints["base"], base_url)
        self.assertEqual(endpoints["api"], api_url)
        self.assertEqual(endpoints["stream"], stream_url)



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import copy
import json
import mock
import requests
import argparse
import logging
import uuid

from tests import base
from st2client import shell
from six.moves import range

LOG = logging.getLogger(__name__)


def _randomize_inquiry_id(inquiry):
    newinquiry = copy.deepcopy(inquiry)
    newinquiry["id"] = str(uuid.uuid4())
    # ID can't have '1440' in it, otherwise our `count()` fails
    # when inspecting the inquiry list output for test:
    # test_list_inquiries_limit()
    while "1440" in newinquiry["id"]:
        newinquiry["id"] = str(uuid.uuid4())
    return newinquiry


def _generate_inquiries(count):
    return [_randomize_inquiry_id(INQUIRY_1) for i in range(count)]


class TestInquiryBase(base.BaseCLITestCase):
    """Base class for "inquiry" CLI tests"""

    capture_output = True

    def __init__(self, *args, **kwargs):
        super(TestInquiryBase, self).__init__(*args, **kwargs)

        self.parser = argparse.ArgumentParser()
        self.parser.add_argument("-t", "--token", dest="token")
        self.parser.add_argument("--api-key", dest="api_key")
        self.shell = shell.Shell()

    def setUp(self):
        super(TestInquiryBase, self).setUp()

    def tearDown(self):
        super(TestInquiryBase, self).tearDown()


SCHEMA_DEFAULT = {
    "title": "response_data",
    "type": "object",
    "properties": {
        "continue": {
            "type": "boolean",
            "description": "Would you like to continue the workflow?",
            "required": True,
        }
    },
}

RESPONSE_DEFAULT = {"continue": True}

SCHEMA_MULTIPLE = {
    "title": "response_data",
    "type": "object",
    "properties": {
        "name": {
            "type": "string",
            "description": "What is your name?",
            "required": True,
        },
        "pin": {
            "type": "integer",
            "description": "What is your PIN?",
            "required": True,
        },
        "paradox": {
            "type": "boolean",
            "description": "This statement is False.",
            "required": True,
        },
    },
}

RESPONSE_MULTIPLE = {"name": "matt", "pin": 1234, "paradox": True}

RESPONSE_BAD = {"foo": "bar"}

INQUIRY_1 = {
    "id": "abcdef",
    "schema": SCHEMA_DEFAULT,
    "roles": [],
    "users": [],
    "route": "",
    "ttl": 1440,
}

INQUIRY_MULTIPLE = {
    "id": "beef",
    "schema": SCHEMA_MULTIPLE,
    "roles": [],
    "users": [],
    "route": "",
    "ttl": 1440,
}


class TestInquirySubcommands(TestInquiryBase):
    @mock.patch.object(
        requests,
        "get",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(INQUIRY_1), 200, "OK")
        ),
    )
    def test_get_inquiry(self):
        """Test retrieval of a single inquiry"""
        inquiry_id = "abcdef"
        args = ["inquiry", "get", inquiry_id]
        retcode = self.shell.run(args)
        self.assertEqual(retcode, 0)

    @mock.patch.object(
        requests,
        "get",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps({}), 404, "NOT FOUND")
        ),
    )
    def test_get_inquiry_not_found(self):
        """Test retrieval of a inquiry that doesn't exist"""
        inquiry_id = "asdbv"
        args = ["inquiry", "get", inquiry_id]
        retcode = self.shell.run(args)
        self.assertEqual(
            'Inquiry "%s" is not found.\n\nERROR: Resource asdbv not found.\n\n'
            % inquiry_id,
            self.stdout.getvalue(),
        )
        self.assertEqual(retcode, 1)

    @mock.patch.object(
        requests,
        "get",
        mock.MagicMock(
            return_value=(
                base.FakeResponse(
                    json.dumps([INQUIRY_1]), 200, "OK", {"X-Total-Count": "1"}
                )
            )
        ),
    )
    def test_list_inquiries(self):
        """Test retrieval of a list of Inquiries"""
        args = ["inquiry", "list"]
        retcode = self.shell.run(args)
        self.assertEqual(retcode, 0)
        self.assertEqual(self.stdout.getvalue().count("1440"), 1)

    @mock.patch.object(
        requests,
        "get",
        mock.MagicMock(
            return_value=(
                base.FakeResponse(
                    json.dumps(_generate_inquiries(50)),
                    200,
                    "OK",
                    {"X-Total-Count": "55"},
                )
            )
        ),
    )
    def test_list_inquiries_limit(self):
        """Test retrieval of a list of Inquiries while using the "limit" option"""
        args = ["inquiry", "list", "-n", "50"]
        retcode = self.shell.run(args)
        self.assertEqual(retcode, 0)
        self.assertEqual(self.stdout.getvalue().count("1440"), 50)
        self.assertIn(
            "Note: Only first 50 inquiries are displayed.", self.stderr.getvalue()
        )

    @mock.patch.object(
        requests,
        "get",
        mock.MagicMock(
            return_value=(
                base.FakeResponse(json.dumps([]), 200, "OK", {"X-Total-Count": "0"})
            )
        ),
    )
    def test_list_empty_inquiries(self):
        """Test empty list of Inquiries"""
        args = ["inquiry", "list"]
        retcode = self.shell.run(args)
        self.assertEqual(retcode, 0)

    @mock.patch.object(
        requests,
        "get",
        mock.MagicMock(
            return_value=(base.FakeResponse(json.dumps(INQUIRY_1), 200, "OK"))
        ),
    )
    @mock.patch.object(
        requests,
        "put",
        mock.MagicMock(
            return_value=(
                base.FakeResponse(
                    json.dumps({"id": "abcdef", "response": RESPONSE_DEFAULT}),
                    200,
                    "OK",
                )
            )
        ),
    )
    @mock.patch("st2client.commands.inquiry.InteractiveForm")
    def test_respond(self, mock_form):
        """Test interactive response"""
        form_instance = mock_form.return_value
        form_instance.initiate_dialog.return_value = RESPONSE_DEFAULT
        args = ["inquiry", "respond", "abcdef"]
        retcode = self.shell.run(args)
        self.assertEqual(retcode, 0)

    @mock.patch.object(
        requests,
        "get",
        mock.MagicMock(
            return_value=(base.FakeResponse(json.dumps(INQUIRY_1), 200, "OK"))
        ),
    )
    @mock.patch.object(
        requests,
        "put",
        mock.MagicMock(
            return_value=(
                base.FakeResponse(
                    json.dumps({"id": "abcdef", "response": RESPONSE_DEFAULT}),
                    200,
                    "OK",
                )
            )
        ),
    )
    def test_respond_response_flag(self):
        """Test response without interactive mode"""
        args = ["inquiry", "respond", "-r", '"%s"' % RESPONSE_DEFAULT, "abcdef"]
        retcode = self.shell.run(args)
        self.assertEqual(retcode, 0)

    @mock.patch.object(
        requests,
        "get",
        mock.MagicMock(
            return_value=(base.FakeResponse(json.dumps(INQUIRY_1), 200, "OK"))
        ),
    )
    @mock.patch.object(
        requests,
        "put",
        mock.MagicMock(
            return_value=(
                base.FakeResponse(json.dumps({}), 400, "400 Client Error: Bad Request")
            )
        ),
    )
    def test_respond_invalid(self):
        """Test invalid response"""
        args = ["inquiry", "respond", "-r", '"%s"' % RESPONSE_BAD, "abcdef"]
        retcode = self.shell.run(args)
        self.assertEqual(retcode, 1)
        self.assertEqual(
            "ERROR: 400 Client Error: Bad Request", self.stdout.getvalue().strip()
        )

    def test_respond_nonexistent_inquiry(self):
        """Test responding to an inquiry that doesn't exist"""
        inquiry_id = "134234"
        args = ["inquiry", "respond", "-r", '"%s"' % RESPONSE_DEFAULT, inquiry_id]
        retcode = self.shell.run(args)
        self.assertEqual(retcode, 1)
        self.assertEqual(
            'ERROR: Resource with id "%s" doesn\'t exist.' % inquiry_id,
            self.stdout.getvalue().strip(),
        )

    @mock.patch.object(
        requests,
        "get",
        mock.MagicMock(
            return_value=(
                base.FakeResponse(json.dumps({}), 404, "404 Client Error: Not Found")
            )
        ),
    )
    @mock.patch("st2client.commands.inquiry.InteractiveForm")
    def test_respond_nonexistent_inquiry_interactive(self, mock_form):
        """Test interactively responding to an inquiry that doesn't exist

        Interactive mode (omitting -r flag) retrieves the inquiry with GET before
        responding with PUT, in order to retrieve the desired schema for this inquiry.
        So, we want to test that interaction separately.
        """
        inquiry_id = "253432"
        form_instance = mock_form.return_value
        form_instance.initiate_dialog.return_value = RESPONSE_DEFAULT
        args = ["inquiry", "respond", inquiry_id]
        retcode = self.shell.run(args)
        self.assertEqual(retcode, 1)
        self.assertEqual(
            'ERROR: Resource with id "%s" doesn\'t exist.' % inquiry_id,
            self.stdout.getvalue().strip(),
        )



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["get_config", "set_config"]

# Stores parsed config dictionary
CONFIG = {}


def get_config():
    """
    Retrieve parsed config object.

    :rtype: ``dict``
    """
    global CONFIG
    return CONFIG


def set_config(config):
    """
    Store parsing config object.

    :type config: ``dict``

    :rtype: ``dict``
    """
    global CONFIG
    CONFIG = config
    return config



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import copy

import unittest2
import six
import mock

from st2client.commands.action import ActionRunCommand
from st2client.models.action import Action, RunnerType


class ActionRunCommandTest(unittest2.TestCase):
    def test_get_params_types(self):
        runner = RunnerType()
        runner_params = {
            "foo": {"immutable": True, "required": True},
            "bar": {"description": "Some param.", "type": "string"},
        }
        runner.runner_parameters = runner_params
        orig_runner_params = copy.deepcopy(runner.runner_parameters)

        action = Action()
        action.parameters = {
            "foo": {"immutable": False},  # Should not be allowed by API.
            "stuff": {"description": "Some param.", "type": "string", "required": True},
        }
        orig_action_params = copy.deepcopy(action.parameters)

        params, rqd, opt, imm = ActionRunCommand._get_params_types(runner, action)
        self.assertEqual(len(list(params.keys())), 3)

        self.assertIn("foo", imm, '"foo" param should be in immutable set.')
        self.assertNotIn("foo", rqd, '"foo" param should not be in required set.')
        self.assertNotIn("foo", opt, '"foo" param should not be in optional set.')

        self.assertIn("bar", opt, '"bar" param should be in optional set.')
        self.assertNotIn("bar", rqd, '"bar" param should not be in required set.')
        self.assertNotIn("bar", imm, '"bar" param should not be in immutable set.')

        self.assertIn("stuff", rqd, '"stuff" param should be in required set.')
        self.assertNotIn("stuff", opt, '"stuff" param should not be in optional set.')
        self.assertNotIn("stuff", imm, '"stuff" param should not be in immutable set.')
        self.assertEqual(
            runner.runner_parameters, orig_runner_params, "Runner params modified."
        )
        self.assertEqual(
            action.parameters, orig_action_params, "Action params modified."
        )

    def test_opt_in_dict_auto_convert(self):
        """Test ability for user to opt-in to dict convert functionality"""

        runner = RunnerType()
        runner.runner_parameters = {}

        action = Action()
        action.ref = "test.action"
        action.parameters = {
            "param_array": {"type": "array"},
        }

        subparser = mock.Mock()
        command = ActionRunCommand(action, self, subparser, name="test")

        mockarg = mock.Mock()
        mockarg.inherit_env = False
        mockarg.parameters = [
            "param_array=foo:bar,foo2:bar2",
        ]

        mockarg.auto_dict = False
        param = command._get_action_parameters_from_args(
            action=action, runner=runner, args=mockarg
        )
        self.assertEqual(param["param_array"], ["foo:bar", "foo2:bar2"])

        mockarg.auto_dict = True
        param = command._get_action_parameters_from_args(
            action=action, runner=runner, args=mockarg
        )
        self.assertEqual(param["param_array"], [{"foo": "bar", "foo2": "bar2"}])

        # set auto_dict back to default
        mockarg.auto_dict = False

    def test_get_params_from_args(self):
        """test_get_params_from_args

        This tests the details of the auto-dict conversion, assuming it's enabled. Please
        see test_opt_in_dict_auto_convert for a test of detecting whether or not this
        functionality is enabled.
        """

        runner = RunnerType()
        runner.runner_parameters = {}

        action = Action()
        action.ref = "test.action"
        action.parameters = {
            "param_string": {"type": "string"},
            "param_integer": {"type": "integer"},
            "param_number": {"type": "number"},
            "param_object": {"type": "object"},
            "param_boolean": {"type": "boolean"},
            "param_array": {"type": "array"},
            "param_array_of_dicts": {
                "type": "array",
                "properties": {
                    "foo": {"type": "string"},
                    "bar": {"type": "integer"},
                    "baz": {"type": "number"},
                    "qux": {"type": "object"},
                    "quux": {"type": "boolean"},
                },
            },
        }

        subparser = mock.Mock()
        command = ActionRunCommand(action, self, subparser, name="test")

        mockarg = mock.Mock()
        mockarg.inherit_env = False
        mockarg.auto_dict = True
        mockarg.parameters = [
            "param_string=hoge",
            "param_integer=123",
            "param_number=1.23",
            "param_object=hoge=1,fuga=2",
            "param_boolean=False",
            "param_array=foo,bar,baz",
            "param_array_of_dicts=foo:HOGE,bar:1,baz:1.23,qux:foo=bar,quux:True",
            "param_array_of_dicts=foo:FUGA,bar:2,baz:2.34,qux:bar=baz,quux:False",
        ]

        param = command._get_action_parameters_from_args(
            action=action, runner=runner, args=mockarg
        )

        self.assertIsInstance(param, dict)
        self.assertEqual(param["param_string"], "hoge")
        self.assertEqual(param["param_integer"], 123)
        self.assertEqual(param["param_number"], 1.23)
        self.assertEqual(param["param_object"], {"hoge": "1", "fuga": "2"})
        self.assertFalse(param["param_boolean"])
        self.assertEqual(param["param_array"], ["foo", "bar", "baz"])

        # checking the result of parsing for array of objects
        self.assertIsInstance(param["param_array_of_dicts"], list)
        self.assertEqual(len(param["param_array_of_dicts"]), 2)
        for param in param["param_array_of_dicts"]:
            self.assertIsInstance(param, dict)
            self.assertIsInstance(param["foo"], str)
            self.assertIsInstance(param["bar"], int)
            self.assertIsInstance(param["baz"], float)
            self.assertIsInstance(param["qux"], dict)
            self.assertIsInstance(param["quux"], bool)

        # set auto_dict back to default
        mockarg.auto_dict = False

    def test_get_params_from_args_read_content_from_file(self):
        runner = RunnerType()
        runner.runner_parameters = {}

        action = Action()
        action.ref = "test.action"
        action.parameters = {
            "param_object": {"type": "object"},
        }

        subparser = mock.Mock()
        command = ActionRunCommand(action, self, subparser, name="test")

        # 1. File doesn't exist
        mockarg = mock.Mock()
        mockarg.inherit_env = False
        mockarg.auto_dict = True
        mockarg.parameters = ["@param_object=doesnt-exist.json"]

        self.assertRaisesRegex(
            ValueError,
            "doesn't exist",
            command._get_action_parameters_from_args,
            action=action,
            runner=runner,
            args=mockarg,
        )

        # 2. Valid file path (we simply read this file)
        mockarg = mock.Mock()
        mockarg.inherit_env = False
        mockarg.auto_dict = True
        mockarg.parameters = ["@param_string=%s" % (__file__)]

        params = command._get_action_parameters_from_args(
            action=action, runner=runner, args=mockarg
        )
        self.assertTrue(isinstance(params["param_string"], six.text_type))
        self.assertTrue(params["param_string"].startswith("# Copyright"))

    def test_get_params_from_args_with_multiple_declarations(self):
        """test_get_params_from_args_with_multiple_declarations

        This tests the details of the auto-dict conversion, assuming it's enabled. Please
        see test_opt_in_dict_auto_convert for a test of detecting whether or not this
        functionality is enabled.
        """

        runner = RunnerType()
        runner.runner_parameters = {}

        action = Action()
        action.ref = "test.action"
        action.parameters = {
            "param_string": {"type": "string"},
            "param_array": {"type": "array"},
            "param_array_of_dicts": {"type": "array"},
        }

        subparser = mock.Mock()
        command = ActionRunCommand(action, self, subparser, name="test")

        mockarg = mock.Mock()
        mockarg.inherit_env = False
        mockarg.auto_dict = True
        mockarg.parameters = [
            "param_string=hoge",  # This value will be overwritten with the next declaration.
            "param_string=fuga",
            "param_array=foo",
            "param_array=bar",
            "param_array_of_dicts=foo:1,bar:2",
            "param_array_of_dicts=hoge:A,fuga:B",
        ]

        param = command._get_action_parameters_from_args(
            action=action, runner=runner, args=mockarg
        )

        # checks to accept multiple declaration only if the array type
        self.assertEqual(param["param_string"], "fuga")
        self.assertEqual(param["param_array"], ["foo", "bar"])
        self.assertEqual(
            param["param_array_of_dicts"],
            [{"foo": "1", "bar": "2"}, {"hoge": "A", "fuga": "B"}],
        )

        # set auto_dict back to default
        mockarg.auto_dict = False

    def test_correctly_process_inherit_env_when_no_parameters_set(self):
        """test_correctly_process_inherit_env_when_no_parameters_set

        This tests that we still correctly pass through the environment variables
        when --inherit-env is set and we run an action that does not have parameters
        """

        runner = RunnerType()
        runner.runner_parameters = {}

        action = Action()
        action.ref = "test.action"

        subparser = mock.Mock()
        command = ActionRunCommand(action, self, subparser, name="test")

        mockarg = mock.Mock()
        mockarg.inherit_env = True
        mockarg.auto_dict = True
        mockarg.parameters = []

        k1 = "key1"
        v1 = "value1"
        k2 = "key2"
        v2 = "value2"

        with mock.patch("os.environ.copy") as mockCopy:
            mockCopy.return_value = {k1: v1, k2: v2}
            param = command._get_action_parameters_from_args(
                action=action, runner=runner, args=mockarg
            )

        self.assertIn("env", param)

        env_params = param["env"]
        self.assertIn(k1, env_params)
        self.assertIn(k2, env_params)
        self.assertEqual(v1, env_params[k1])
        self.assertEqual(v2, env_params[k2])

    def test_correctly_process_inherit_env_when_parameters_set(self):
        """test_correctly_process_inherit_env_when_parameters_set

        This tests that we still correctly pass through the environment variables
        when --inherit-env is set and we run an action that has action parameters set
        """

        runner = RunnerType()
        runner.runner_parameters = {}

        action = Action()
        action.ref = "test.action"
        action.parameters = {
            "param_string": {"type": "string"},
            "param_array": {"type": "array"},
            "param_array_of_dicts": {"type": "array"},
        }

        subparser = mock.Mock()
        command = ActionRunCommand(action, self, subparser, name="test")

        p_string = "param_string"
        p_array = "param_array"
        p_ra_dicts = "param_array_of_dicts"
        mockarg = mock.Mock()
        mockarg.inherit_env = True
        mockarg.auto_dict = True
        mockarg.parameters = [
            f"{p_string}=hoge",
            f"{p_array}=foo,bar",
            f"{p_ra_dicts}=foo:1,bar:2",
        ]

        k1 = "key1"
        v1 = "value1"
        k2 = "key2"
        v2 = "value2"

        with mock.patch("os.environ.copy") as mockCopy:
            mockCopy.return_value = {k1: v1, k2: v2}
            param = command._get_action_parameters_from_args(
                action=action, runner=runner, args=mockarg
            )

        self.assertIn("env", param)

        env_params = param["env"]
        self.assertIn(k1, env_params)
        self.assertIn(k2, env_params)
        self.assertEqual(v1, env_params[k1])
        self.assertEqual(v2, env_params[k2])
        self.assertIn(p_string, param)
        self.assertEqual("hoge", param[p_string])
        self.assertIn(p_array, param)
        self.assertIn("foo", param[p_array])
        self.assertIn("bar", param[p_array])
        self.assertIn(p_ra_dicts, param)
        self.assertDictEqual({"foo": "1", "bar": "2"}, param[p_ra_dicts][0])

    def test_correctly_generate_empty_params_no_inherit_empty_parameters(self):
        """test_correctly_generate_empty_params_no_inherit_empty_parameters

        Verifies that we return an empty dict when we do not provide inherit env and parameters
        """

        runner = RunnerType()
        runner.runner_parameters = {}

        action = Action()
        action.ref = "test.action"

        subparser = mock.Mock()
        command = ActionRunCommand(action, self, subparser, name="test")

        mockarg = mock.Mock()
        mockarg.inherit_env = False
        mockarg.auto_dict = True
        mockarg.parameters = []

        k1 = "key1"
        v1 = "value1"
        k2 = "key2"
        v2 = "value2"

        with mock.patch("os.environ.copy") as mockCopy:
            mockCopy.return_value = {k1: v1, k2: v2}
            param = command._get_action_parameters_from_args(
                action=action, runner=runner, args=mockarg
            )

        self.assertDictEqual({}, param)



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import os
import logging
import warnings

import six

from st2client import models
from st2client.utils import httpclient
from st2client.models.core import ResourceManager
from st2client.models.core import ActionAliasResourceManager
from st2client.models.core import ActionAliasExecutionManager
from st2client.models.core import ActionResourceManager
from st2client.models.core import ExecutionResourceManager
from st2client.models.core import InquiryResourceManager
from st2client.models.core import TriggerInstanceResourceManager
from st2client.models.core import PackResourceManager
from st2client.models.core import ConfigManager
from st2client.models.core import WebhookManager
from st2client.models.core import StreamManager
from st2client.models.core import WorkflowManager
from st2client.models.core import ServiceRegistryGroupsManager
from st2client.models.core import ServiceRegistryMembersManager
from st2client.models.core import add_auth_token_to_kwargs_from_env
from st2client.models.core import KeyValuePairResourceManager


LOG = logging.getLogger(__name__)

# Default values for the options not explicitly specified by the user
DEFAULT_API_PORT = 9101
DEFAULT_AUTH_PORT = 9100
DEFAULT_STREAM_PORT = 9102

DEFAULT_BASE_URL = "http://127.0.0.1"
DEFAULT_API_VERSION = "v1"


class Client(object):
    def __init__(
        self,
        base_url=None,
        auth_url=None,
        api_url=None,
        stream_url=None,
        api_version=None,
        cacert=None,
        debug=False,
        token=None,
        api_key=None,
        basic_auth=None,
    ):
        # Get CLI options. If not given, then try to get it from the environment.
        self.endpoints = dict()

        # Populate the endpoints
        if base_url:
            self.endpoints["base"] = base_url
        else:
            self.endpoints["base"] = os.environ.get("ST2_BASE_URL", DEFAULT_BASE_URL)

        api_version = api_version or os.environ.get(
            "ST2_API_VERSION", DEFAULT_API_VERSION
        )

        self.endpoints["exp"] = "%s:%s/%s" % (
            self.endpoints["base"],
            DEFAULT_API_PORT,
            "exp",
        )

        if api_url:
            self.endpoints["api"] = api_url
        else:
            self.endpoints["api"] = os.environ.get(
                "ST2_API_URL",
                "%s:%s/%s" % (self.endpoints["base"], DEFAULT_API_PORT, api_version),
            )

        if auth_url:
            self.endpoints["auth"] = auth_url
        else:
            self.endpoints["auth"] = os.environ.get(
                "ST2_AUTH_URL", "%s:%s" % (self.endpoints["base"], DEFAULT_AUTH_PORT)
            )

        if stream_url:
            self.endpoints["stream"] = stream_url
        else:
            self.endpoints["stream"] = os.environ.get(
                "ST2_STREAM_URL",
                "%s:%s/%s" % (self.endpoints["base"], DEFAULT_STREAM_PORT, api_version),
            )

        if cacert is not None:
            self.cacert = cacert
        else:
            self.cacert = os.environ.get("ST2_CACERT", None)

        # Note: boolean is also a valid value for "cacert"
        is_cacert_string = isinstance(self.cacert, six.string_types)
        if self.cacert and is_cacert_string and not os.path.isfile(self.cacert):
            raise ValueError('CA cert file "%s" does not exist.' % (self.cacert))

        self.debug = debug

        # Note: This is a nasty hack for now, but we need to get rid of the decrator abuse
        if token:
            os.environ["ST2_AUTH_TOKEN"] = token

        self.token = token

        if api_key:
            os.environ["ST2_API_KEY"] = api_key

        self.api_key = api_key

        if basic_auth:
            # NOTE: We assume username can't contain colons
            if len(basic_auth.split(":", 1)) != 2:
                raise ValueError(
                    "basic_auth config options needs to be in the "
                    "username:password notation"
                )

            self.basic_auth = tuple(basic_auth.split(":", 1))
        else:
            self.basic_auth = None

        # Instantiate resource managers and assign appropriate API endpoint.
        self.managers = dict()
        self.managers["Token"] = ResourceManager(
            models.Token,
            self.endpoints["auth"],
            cacert=self.cacert,
            debug=self.debug,
            basic_auth=self.basic_auth,
        )
        self.managers["RunnerType"] = ResourceManager(
            models.RunnerType,
            self.endpoints["api"],
            cacert=self.cacert,
            debug=self.debug,
            basic_auth=self.basic_auth,
        )
        self.managers["Action"] = ActionResourceManager(
            models.Action,
            self.endpoints["api"],
            cacert=self.cacert,
            debug=self.debug,
            basic_auth=self.basic_auth,
        )
        self.managers["ActionAlias"] = ActionAliasResourceManager(
            models.ActionAlias,
            self.endpoints["api"],
            cacert=self.cacert,
            debug=self.debug,
            basic_auth=self.basic_auth,
        )
        self.managers["ActionAliasExecution"] = ActionAliasExecutionManager(
            models.ActionAliasExecution,
            self.endpoints["api"],
            cacert=self.cacert,
            debug=self.debug,
            basic_auth=self.basic_auth,
        )
        self.managers["ApiKey"] = ResourceManager(
            models.ApiKey,
            self.endpoints["api"],
            cacert=self.cacert,
            debug=self.debug,
            basic_auth=self.basic_auth,
        )
        self.managers["Config"] = ConfigManager(
            models.Config,
            self.endpoints["api"],
            cacert=self.cacert,
            debug=self.debug,
            basic_auth=self.basic_auth,
        )
        self.managers["ConfigSchema"] = ResourceManager(
            models.ConfigSchema,
            self.endpoints["api"],
            cacert=self.cacert,
            debug=self.debug,
            basic_auth=self.basic_auth,
        )
        self.managers["Execution"] = ExecutionResourceManager(
            models.Execution,
            self.endpoints["api"],
            cacert=self.cacert,
            debug=self.debug,
            basic_auth=self.basic_auth,
        )
        # NOTE: LiveAction has been deprecated in favor of Execution. It will be left here for
        # backward compatibility reasons until v3.2.0
        self.managers["LiveAction"] = self.managers["Execution"]
        self.managers["Inquiry"] = InquiryResourceManager(
            models.Inquiry,
            self.endpoints["api"],
            cacert=self.cacert,
            debug=self.debug,
            basic_auth=self.basic_auth,
        )
        self.managers["Pack"] = PackResourceManager(
            models.Pack,
            self.endpoints["api"],
            cacert=self.cacert,
            debug=self.debug,
            basic_auth=self.basic_auth,
        )
        self.managers["Policy"] = ResourceManager(
            models.Policy,
            self.endpoints["api"],
            cacert=self.cacert,
            debug=self.debug,
            basic_auth=self.basic_auth,
        )
        self.managers["PolicyType"] = ResourceManager(
            models.PolicyType,
            self.endpoints["api"],
            cacert=self.cacert,
            debug=self.debug,
            basic_auth=self.basic_auth,
        )
        self.managers["Rule"] = ResourceManager(
            models.Rule,
            self.endpoints["api"],
            cacert=self.cacert,
            debug=self.debug,
            basic_auth=self.basic_auth,
        )
        self.managers["Sensor"] = ResourceManager(
            models.Sensor,
            self.endpoints["api"],
            cacert=self.cacert,
            debug=self.debug,
            basic_auth=self.basic_auth,
        )
        self.managers["TriggerType"] = ResourceManager(
            models.TriggerType,
            self.endpoints["api"],
            cacert=self.cacert,
            debug=self.debug,
            basic_auth=self.basic_auth,
        )
        self.managers["Trigger"] = ResourceManager(
            models.Trigger,
            self.endpoints["api"],
            cacert=self.cacert,
            debug=self.debug,
            basic_auth=self.basic_auth,
        )
        self.managers["TriggerInstance"] = TriggerInstanceResourceManager(
            models.TriggerInstance,
            self.endpoints["api"],
            cacert=self.cacert,
            debug=self.debug,
            basic_auth=self.basic_auth,
        )
        self.managers["KeyValuePair"] = KeyValuePairResourceManager(
            models.KeyValuePair,
            self.endpoints["api"],
            cacert=self.cacert,
            debug=self.debug,
            basic_auth=self.basic_auth,
        )
        self.managers["Webhook"] = WebhookManager(
            models.Webhook,
            self.endpoints["api"],
            cacert=self.cacert,
            debug=self.debug,
            basic_auth=self.basic_auth,
        )
        self.managers["Timer"] = ResourceManager(
            models.Timer,
            self.endpoints["api"],
            cacert=self.cacert,
            debug=self.debug,
            basic_auth=self.basic_auth,
        )
        self.managers["Trace"] = ResourceManager(
            models.Trace,
            self.endpoints["api"],
            cacert=self.cacert,
            debug=self.debug,
            basic_auth=self.basic_auth,
        )
        self.managers["RuleEnforcement"] = ResourceManager(
            models.RuleEnforcement,
            self.endpoints["api"],
            cacert=self.cacert,
            debug=self.debug,
            basic_auth=self.basic_auth,
        )
        self.managers["Stream"] = StreamManager(
            self.endpoints["stream"],
            cacert=self.cacert,
            debug=self.debug,
            basic_auth=self.basic_auth,
        )
        self.managers["Workflow"] = WorkflowManager(
            self.endpoints["api"],
            cacert=self.cacert,
            debug=self.debug,
            basic_auth=self.basic_auth,
        )

        # Service Registry
        self.managers["ServiceRegistryGroups"] = ServiceRegistryGroupsManager(
            models.ServiceRegistryGroup,
            self.endpoints["api"],
            cacert=self.cacert,
            debug=self.debug,
            basic_auth=self.basic_auth,
        )

        self.managers["ServiceRegistryMembers"] = ServiceRegistryMembersManager(
            models.ServiceRegistryMember,
            self.endpoints["api"],
            cacert=self.cacert,
            debug=self.debug,
            basic_auth=self.basic_auth,
        )

        # RBAC
        self.managers["Role"] = ResourceManager(
            models.Role,
            self.endpoints["api"],
            cacert=self.cacert,
            debug=self.debug,
            basic_auth=self.basic_auth,
        )
        self.managers["UserRoleAssignment"] = ResourceManager(
            models.UserRoleAssignment,
            self.endpoints["api"],
            cacert=self.cacert,
            debug=self.debug,
            basic_auth=self.basic_auth,
        )

    @add_auth_token_to_kwargs_from_env
    def get_user_info(self, **kwargs):
        """
        Retrieve information about the current user which is authenticated against StackStorm API.

        :rtype: ``dict``
        """
        url = "/user"
        client = httpclient.HTTPClient(
            root=self.endpoints["api"],
            cacert=self.cacert,
            debug=self.debug,
            basic_auth=self.basic_auth,
        )
        response = client.get(url=url, **kwargs)

        if response.status_code != 200:
            ResourceManager.handle_error(response)

        return response.json()

    @property
    def actions(self):
        return self.managers["Action"]

    @property
    def apikeys(self):
        return self.managers["ApiKey"]

    @property
    def keys(self):
        return self.managers["KeyValuePair"]

    @property
    def executions(self):
        return self.managers["Execution"]

    # NOTE: LiveAction has been deprecated in favor of Execution. It will be left here for
    # backward compatibility reasons until v3.2.0
    @property
    def liveactions(self):
        warnings.warn(
            (
                "st2client.liveactions has been renamed to st2client.executions, please "
                "update your code"
            ),
            DeprecationWarning,
        )
        return self.executions

    @property
    def inquiries(self):
        return self.managers["Inquiry"]

    @property
    def packs(self):
        return self.managers["Pack"]

    @property
    def policies(self):
        return self.managers["Policy"]

    @property
    def policytypes(self):
        return self.managers["PolicyType"]

    @property
    def rules(self):
        return self.managers["Rule"]

    @property
    def runners(self):
        return self.managers["RunnerType"]

    @property
    def sensors(self):
        return self.managers["Sensor"]

    @property
    def tokens(self):
        return self.managers["Token"]

    @property
    def triggertypes(self):
        return self.managers["TriggerType"]

    @property
    def triggerinstances(self):
        return self.managers["TriggerInstance"]

    @property
    def trace(self):
        return self.managers["Trace"]

    @property
    def ruleenforcements(self):
        return self.managers["RuleEnforcement"]

    @property
    def webhooks(self):
        return self.managers["Webhook"]

    @property
    def workflows(self):
        return self.managers["Workflow"]



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import absolute_import

import abc
import six
import logging

from st2client.formatters import doc


LOG = logging.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class Branch(object):
    """Represents a branch of related commands in the command tree."""

    def __init__(self, name, description, app, subparsers, parent_parser=None):
        self.name = name
        self.description = description
        self.app = app
        self.parent_parser = parent_parser
        self.parser = subparsers.add_parser(
            self.name, description=self.description, help=self.description
        )
        self.commands = dict()


@six.add_metaclass(abc.ABCMeta)
class Command(object):
    """Represents a commandlet in the command tree."""

    def __init__(
        self, name, description, app, subparsers, parent_parser=None, add_help=True
    ):
        self.name = name
        self.description = description
        self.app = app
        self.parent_parser = parent_parser
        self.parser = subparsers.add_parser(
            self.name,
            description=self.description,
            help=self.description,
            add_help=add_help,
        )
        self.parser.set_defaults(func=self.run_and_print)

    @abc.abstractmethod
    def run(self, args, **kwargs):
        """
        This method should be invoked from run_and_print. The separation of run
        is to let the core logic be testable.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def run_and_print(self, args, **kwargs):
        """
        This method is invoked when the corresponding command is executed
        from the command line.
        """
        raise NotImplementedError

    def format_output(self, subject, formatter, *args, **kwargs):
        json = kwargs.get("json", False)
        yaml = kwargs.get("yaml", False)

        if json:
            func = doc.JsonFormatter.format
        elif yaml:
            func = doc.YAMLFormatter.format
        else:
            func = formatter.format
        return func(subject, *args, **kwargs)

    def print_output(self, subject, formatter, *args, **kwargs):
        if subject:
            output = self.format_output(subject, formatter, *args, **kwargs)
            print(output)
        else:
            print("No matching items found")



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import six
import os
import sys
import mock
import json
import logging
import tempfile
import unittest2

from io import BytesIO
from six.moves import StringIO

import pytest

from tests import base
from tests.fixtures import loader

from st2client import shell
from st2client.formatters import table
from st2client.utils import jsutil
from st2client.utils import httpclient
from st2client.utils import color


LOG = logging.getLogger(__name__)

FIXTURES_MANIFEST = {
    "executions": [
        "execution.json",
        "execution_result_has_carriage_return.json",
        "execution_unicode.json",
        "execution_double_backslash.json",
        "execution_with_stack_trace.json",
        "execution_with_schema.json",
    ],
    "results": [
        "execution_get_default.txt",
        "execution_get_detail.txt",
        "execution_get_result_by_key.txt",
        "execution_result_has_carriage_return.txt",
        "execution_result_has_carriage_return_py3.txt",
        "execution_get_attributes.txt",
        "execution_list_attr_start_timestamp.txt",
        "execution_list_empty_response_start_timestamp_attr.txt",
        "execution_unescape_newline.txt",
        "execution_unicode.txt",
        "execution_double_backslash.txt",
        "execution_unicode_py3.txt",
        "execution_get_has_schema.txt",
    ],
}

FIXTURES = loader.load_fixtures(fixtures_dict=FIXTURES_MANIFEST)
EXECUTION = FIXTURES["executions"]["execution.json"]
UNICODE = FIXTURES["executions"]["execution_unicode.json"]
DOUBLE_BACKSLASH = FIXTURES["executions"]["execution_double_backslash.json"]
OUTPUT_SCHEMA = FIXTURES["executions"]["execution_with_schema.json"]
NEWLINE = FIXTURES["executions"]["execution_with_stack_trace.json"]
HAS_CARRIAGE_RETURN = FIXTURES["executions"][
    "execution_result_has_carriage_return.json"
]


@pytest.fixture(scope="class")
def redirect_console_for_pytest(request):
    request.cls._redirect_console = lambda *args: None
    request.cls._undo_console_redirect = lambda *args: None


@pytest.mark.usefixtures("redirect_console_for_pytest")
class TestExecutionResultFormatter(unittest2.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestExecutionResultFormatter, self).__init__(*args, **kwargs)
        self.shell = shell.Shell()
        self.table = table.SingleRowTable()
        color.DISABLED = True
        self._capsys = None

    def setUp(self):
        self.fd, self.path = tempfile.mkstemp()
        self._redirect_console(self.path)
        self.maxDiff = None

    def tearDown(self):
        self._undo_console_redirect()
        os.close(self.fd)
        os.unlink(self.path)

    def _redirect_console(self, path):
        sys.stdout = open(path, "w")
        sys.stderr = open(path, "w")

    def _undo_console_redirect(self):
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__

    @pytest.fixture(autouse=True)
    def redirect_console_with_capsys(self, capsys):
        self._capsys = capsys
        yield
        self._capsys = None

    def _get_output(self):
        if self._capsys:
            captured = self._capsys.readouterr()
            # stderr actually has some extra newlines not accounted for in old capture method.
            content = captured.out  # + captured.err
            return content
        self._undo_console_redirect()
        with open(self.path, "r") as fd:
            content = fd.read()
        return content

    def test_console_redirect(self):
        message = "Hello, World!"
        print(message)
        content = self._get_output()
        self.assertEqual(content.replace("\n", ""), message)

    def test_execution_get_default(self):
        argv = ["execution", "get", EXECUTION["id"]]
        content = self._get_execution(argv)
        self.assertEqual(content, FIXTURES["results"]["execution_get_default.txt"])

    def test_execution_get_attributes(self):
        argv = [
            "execution",
            "get",
            EXECUTION["id"],
            "--attr",
            "status",
            "end_timestamp",
        ]
        content = self._get_execution(argv)
        self.assertEqual(content, FIXTURES["results"]["execution_get_attributes.txt"])

    def test_execution_get_default_in_json(self):
        argv = ["execution", "get", EXECUTION["id"], "-j"]
        content = self._get_execution(argv)
        self.assertEqual(
            json.loads(content),
            jsutil.get_kvps(
                EXECUTION,
                [
                    "id",
                    "action.ref",
                    "context.user",
                    "start_timestamp",
                    "end_timestamp",
                    "status",
                    "parameters",
                    "result",
                ],
            ),
        )

    def test_execution_get_detail(self):
        argv = ["execution", "get", EXECUTION["id"], "-d"]
        content = self._get_execution(argv)
        self.assertEqual(content, FIXTURES["results"]["execution_get_detail.txt"])

    def test_execution_with_schema(self):
        argv = ["execution", "get", OUTPUT_SCHEMA["id"]]
        content = self._get_schema_execution(argv)
        self.assertEqual(content, FIXTURES["results"]["execution_get_has_schema.txt"])

    @mock.patch.object(
        httpclient.HTTPClient,
        "get",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(NEWLINE), 200, "OK", {})
        ),
    )
    def test_execution_unescape_newline(self):
        """Ensure client renders newline characters"""

        argv = ["execution", "get", NEWLINE["id"]]
        self.assertEqual(self.shell.run(argv), 0)
        content = self._get_output()

        # NOTE: For some reason CI and locally the indent is different sometimes (2 vs 4 spaces)
        # even though it's using the same code
        expected = FIXTURES["results"]["execution_unescape_newline.txt"]
        expected = expected.replace("      '", "  '").replace("  '", "'")
        content = content.replace("      '", "  '").replace("  '", "'")
        content = content.replace("      '", "  '")
        self.assertEqual(content, expected)

    @mock.patch.object(
        httpclient.HTTPClient,
        "get",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(UNICODE), 200, "OK", {})
        ),
    )
    def test_execution_unicode(self):
        """Ensure client renders unicode escape sequences"""

        argv = ["execution", "get", UNICODE["id"]]
        self.assertEqual(self.shell.run(argv), 0)
        content = self._get_output()

        content = content.replace(r"\xE2\x80\xA1", r"\u2021")
        self.assertEqual(content, FIXTURES["results"]["execution_unicode_py3.txt"])

    @mock.patch.object(
        httpclient.HTTPClient,
        "get",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(DOUBLE_BACKSLASH), 200, "OK", {})
        ),
    )
    def test_execution_double_backslash_not_unicode_escape_sequence(self):
        argv = ["execution", "get", DOUBLE_BACKSLASH["id"]]
        self.assertEqual(self.shell.run(argv), 0)
        content = self._get_output()

        self.assertEqual(content, FIXTURES["results"]["execution_double_backslash.txt"])

    def test_execution_get_detail_in_json(self):
        argv = ["execution", "get", EXECUTION["id"], "-d", "-j"]
        content = self._get_execution(argv)
        content_dict = json.loads(content)
        # Sufficient to check if output contains all expected keys. The entire result will not
        # match as content will contain characters which improve rendering.
        for k in six.iterkeys(EXECUTION):
            if k in ["liveaction", "callback"]:
                continue

            if k in content:
                continue
            self.assertTrue(
                False, "Missing key %s. %s != %s" % (k, EXECUTION, content_dict)
            )

    def test_execution_get_result_by_key(self):
        argv = ["execution", "get", EXECUTION["id"], "-k", "localhost.stdout"]
        content = self._get_execution(argv)
        self.assertEqual(
            content, FIXTURES["results"]["execution_get_result_by_key.txt"]
        )

    def test_execution_get_result_by_key_in_json(self):
        argv = ["execution", "get", EXECUTION["id"], "-k", "localhost.stdout", "-j"]
        content = self._get_execution(argv)
        self.assertDictEqual(
            json.loads(content), jsutil.get_kvps(EXECUTION, ["result.localhost.stdout"])
        )

    @mock.patch.object(
        httpclient.HTTPClient,
        "get",
        mock.MagicMock(
            return_value=base.FakeResponse(
                json.dumps(HAS_CARRIAGE_RETURN), 200, "OK", {}
            )
        ),
    )
    def test_execution_get_detail_with_carriage_return(self):
        argv = ["execution", "get", HAS_CARRIAGE_RETURN["id"], "-d"]
        self.assertEqual(self.shell.run(argv), 0)
        content = self._get_output()

        self.assertEqual(
            content, FIXTURES["results"]["execution_result_has_carriage_return_py3.txt"]
        )

    @mock.patch.object(
        httpclient.HTTPClient,
        "get",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps([EXECUTION]), 200, "OK", {})
        ),
    )
    def test_execution_list_attribute_provided(self):
        # Client shouldn't throw if "-a" flag is provided when listing executions
        argv = ["execution", "list", "-a", "start_timestamp"]
        self.assertEqual(self.shell.run(argv), 0)
        content = self._get_output()

        self.assertEqual(
            content, FIXTURES["results"]["execution_list_attr_start_timestamp.txt"]
        )

    @mock.patch.object(
        httpclient.HTTPClient,
        "get",
        mock.MagicMock(return_value=base.FakeResponse(json.dumps([]), 200, "OK", {})),
    )
    def test_execution_list_attribute_provided_empty_response(self):
        # Client shouldn't throw if "-a" flag is provided, but there are no executions
        argv = ["execution", "list", "-a", "start_timestamp"]
        self.assertEqual(self.shell.run(argv), 0)
        content = self._get_output()

        self.assertEqual(
            content,
            FIXTURES["results"][
                "execution_list_empty_response_start_timestamp_attr.txt"
            ],
        )

    @mock.patch.object(
        httpclient.HTTPClient,
        "get",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(EXECUTION), 200, "OK", {})
        ),
    )
    def _get_execution(self, argv):
        self.assertEqual(self.shell.run(argv), 0)
        content = self._get_output()

        return content

    @mock.patch.object(
        httpclient.HTTPClient,
        "get",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(OUTPUT_SCHEMA), 200, "OK", {})
        ),
    )
    def _get_schema_execution(self, argv):
        self.assertEqual(self.shell.run(argv), 0)
        content = self._get_output()

        return content

    def test_SinlgeRowTable_notebox_one(self):
        with mock.patch("sys.stderr", new=StringIO()) as fackety_fake:
            expected = (
                "Note: Only one action execution is displayed. Use -n/--last flag for "
                "more results."
            )
            print(self.table.note_box("action executions", 1))
            content = fackety_fake.getvalue().split("|")[1].strip()
            self.assertEqual(content, expected)

    def test_SinlgeRowTable_notebox_zero(self):
        with mock.patch("sys.stderr", new=BytesIO()) as fackety_fake:
            contents = fackety_fake.getvalue()
            self.assertEqual(contents, b"")

    def test_SinlgeRowTable_notebox_default(self):
        with mock.patch("sys.stderr", new=StringIO()) as fackety_fake:
            expected = (
                "Note: Only first 50 action executions are displayed. Use -n/--last flag "
                "for more results."
            )
            print(self.table.note_box("action executions", 50))
            content = fackety_fake.getvalue().split("|")[1].strip()
            self.assertEqual(content, expected)
        with mock.patch("sys.stderr", new=StringIO()) as fackety_fake:
            expected = (
                "Note: Only first 15 action executions are displayed. Use -n/--last flag "
                "for more results."
            )
            print(self.table.note_box("action executions", 15))
            content = fackety_fake.getvalue().split("|")[1].strip()
            self.assertEqual(content, expected)



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

try:
    import simplejson as json
except ImportError:
    import json

import os
import six
import yaml


ALLOWED_EXTS = [".json", ".yaml", ".yml", ".txt"]
PARSER_FUNCS = {".json": json.load, ".yml": yaml.safe_load, ".yaml": yaml.safe_load}


def get_fixtures_base_path():
    return os.path.dirname(__file__)


def load_content(file_path):
    """
    Loads content from file_path if file_path's extension
    is one of allowed ones (See ALLOWED_EXTS).
    Throws UnsupportedMetaException on disallowed filetypes.
    :param file_path: Absolute path to the file to load content from.
    :type file_path: ``str``
    :rtype: ``dict``
    """
    file_name, file_ext = os.path.splitext(file_path)

    if file_ext not in ALLOWED_EXTS:
        raise Exception(
            "Unsupported meta type %s, file %s. Allowed: %s"
            % (file_ext, file_path, ALLOWED_EXTS)
        )

    parser_func = PARSER_FUNCS.get(file_ext, None)

    with open(file_path, "r") as fd:
        return parser_func(fd) if parser_func else fd.read()


def load_fixtures(fixtures_dict=None):
    """
    Loads fixtures specified in fixtures_dict. This method must be
    used for fixtures that don't have associated data models. We
    simply want to load the meta into dict objects.
    fixtures_dict should be of the form:
    {
        'actionchains': ['actionchain1.json', 'actionchain2.json'],
        'workflows': ['workflow.yaml']
    }
    :param fixtures_dict: Dictionary specifying the fixtures to load for each type.
    :type fixtures_dict: ``dict``
    :rtype: ``dict``
    """
    if fixtures_dict is None:
        fixtures_dict = {}

    all_fixtures = {}
    fixtures_base_path = get_fixtures_base_path()
    for fixture_type, fixtures in six.iteritems(fixtures_dict):
        loaded_fixtures = {}
        for fixture in fixtures:
            fixture_path = fixtures_base_path + "/" + fixture
            fixture_dict = load_content(fixture_path)
            loaded_fixtures[fixture] = fixture_dict
        all_fixtures[fixture_type] = loaded_fixtures

    return all_fixtures
